# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
import sys
from abc import ABC, abstractmethod
from typing import Any, List, Optional, cast, Dict, Union

from pygelf import gelf, GelfTcpHandler, GelfUdpHandler, GelfTlsHandler, GelfHttpHandler  # type: ignore[import-untyped]

VERBOSE = 5
gelf.LEVELS.update({VERBOSE: 8})

DEFAULT_FORMAT = '%(asctime)s %(name)s %(levelname)s %(message)s'

BLACK, RED, GREEN, YELLOW, BLUE, MAGENTA, CYAN, WHITE = range(8)

RESET_SEQ = "\033[0m"
COLOR_SEQ = "\033[1;%dm"
BOLD_SEQ = "\033[1m"

COLORS = {
    'RESET': RESET_SEQ,
    'BOLD': BOLD_SEQ,
    'BLACK': COLOR_SEQ % (30 + BLACK),
    'RED': COLOR_SEQ % (30 + RED),
    'GREEN': COLOR_SEQ % (30 + GREEN),
    'YELLOW': COLOR_SEQ % (30 + YELLOW),
    'BLUE': COLOR_SEQ % (30 + BLUE),
    'MAGENTA': COLOR_SEQ % (30 + MAGENTA),
    'CYAN': COLOR_SEQ % (30 + CYAN),
    'WHITE': COLOR_SEQ % (30 + WHITE),
}

COLORS_SCHEME = {
    'WARNING': 'YELLOW',
    'VERBOSE': 'BLUE',
    'INFO': 'GREEN',
    'ERROR': 'RED',
    'CRITICAL': 'RED'
}

LOGGING_LEVELS_BY_NAME = {
    '5': VERBOSE,
    'verbose': VERBOSE,
    '10': logging.DEBUG,
    'debug': logging.DEBUG,
    '20': logging.INFO,
    'info': logging.INFO,
    '30': logging.WARNING,
    'warn': logging.WARNING,
    'warning': logging.WARNING,
    '40': logging.ERROR,
    'error': logging.ERROR,
    '50': logging.CRITICAL,
    'critical': logging.CRITICAL
}

LOGGING_NAMES_BY_LEVEL = {
    VERBOSE: 'verbose',
    logging.DEBUG: 'debug',
    logging.INFO: 'info',
    logging.WARNING: 'warning',
    logging.ERROR: 'error',
    logging.CRITICAL: 'critical'
}


class VerboseLogger(ABC):
    @abstractmethod
    def verbose(self, msg: object, *args: object, **kwargs: Any) -> None:
        pass


class EnhancedLogger(logging.Logger, VerboseLogger):
    def __init__(self, name: str, level: int = logging.NOTSET):
        super().__init__(name, level)
        logging.addLevelName(VERBOSE, 'VERBOSE')

    def verbose(self, msg: object, *args: object, **kwargs: Any) -> None:
        if self.isEnabledFor(VERBOSE):
            if sys.version_info[:2] >= (3, 11):
                # https://github.com/python/cpython/pull/28287
                self._log(VERBOSE, msg, args, **kwargs, stacklevel=2)
            else:
                self._log(VERBOSE, msg, args, **kwargs)

    def makeRecord(self, name: str, level: int, fn: str, lno: int, msg: object, args,  # type: ignore[no-untyped-def]
                   exc_info, func=None, extra=None,
                   sinfo=None) -> logging.LogRecord:
        record = super().makeRecord(name, level, fn, lno, msg, args, exc_info, func, extra, sinfo)
        caller = record.__dict__.get('real_caller')
        if caller is not None:
            record.__dict__.update(record.__dict__.pop('real_caller'))

        return record


class EnhancedLogRecord(logging.LogRecord):
    def getMessage(self) -> str:
        message = super().getMessage()
        prefix = self.__dict__.get('prefix')
        if prefix is not None:
            message = prefix + message
        return message


logging.setLoggerClass(EnhancedLogger)
logging.setLogRecordFactory(EnhancedLogRecord)


class LogFormatter(logging.Formatter):
    def __init__(self, fmt: str = None, datefmt: str = None,
                 colorize: bool = False, correct_newlines: bool = False):
        super().__init__(fmt, datefmt)
        self.colorize = colorize
        self.correct_newlines = correct_newlines

    def _format(self, record: logging.LogRecord) -> str:
        s = super().format(record)
        if self.colorize and record.levelname in COLORS_SCHEME:
            s = COLORS[COLORS_SCHEME[record.levelname]] + s + COLORS['RESET']
        return s

    def format(self, record: logging.LogRecord) -> str:
        if self.correct_newlines:
            messages = str(record.msg).split('\n')
            if len(messages) == 1:
                return self._format(record)
        else:
            return self._format(record)

        orig_msg = record.msg
        try:
            s = ''
            for message in messages:
                if s != '':
                    s += '\n'
                record.msg = message
                s += self._format(record)
            return s
        finally:
            record.msg = orig_msg


class StdoutHandler(logging.StreamHandler):
    def __init__(self, formatter: LogFormatter):
        super().__init__(sys.stdout)
        self.formatter: LogFormatter = formatter

    def emit(self, record: logging.LogRecord) -> None:
        if 'ignore_stdout' in record.__dict__:
            return
        super().emit(record)
        # TODO: if output stuck, then add here self.flush()
        # More about, see at https://stackoverflow.com/questions/16633911


class FileHandlerWithHeader(logging.FileHandler):
    def __init__(self, formatter: LogFormatter, filename: str, header: str = None,
                 mode: str = 'a', encoding: str = None):
        # Store the header information.
        self.header = header

        # Call the parent __init__
        logging.FileHandler.__init__(self, filename, mode, encoding)

        # Write the header if it was specified
        if header:
            self.stream.write('%s\n' % header)

        self.formatter: LogFormatter = formatter

    def emit(self, record: logging.LogRecord) -> None:
        # Call the parent class emit function.
        logging.FileHandler.emit(self, record)


class LogHandler:

    def __init__(self,
                 target: str,
                 level: str,
                 colorize: bool = False,
                 correct_newlines: bool = False,
                 filemode: str = 'a',
                 format: str = DEFAULT_FORMAT,  # pylint: disable=redefined-builtin
                 datefmt: str = None,
                 header: str = None,
                 **kwargs: Union[str, bool, int]):

        self._colorize = colorize
        self._correct_newlines = correct_newlines
        self._format = format
        self._datefmt = datefmt
        self._header = header

        self._formatter = LogFormatter(self._format, self._datefmt,
                                       colorize=self._colorize,
                                       correct_newlines=self._correct_newlines)

        self.handler: logging.Handler
        if target.lower() == 'stdout':
            self._target = 'stdout'
            self.handler = StdoutHandler(self._formatter)
        elif target.lower() == 'graylog':
            self._target = 'graylog'
            if not kwargs.get('host'):
                raise Exception('Graylog host is not defined')
            if not kwargs.get('port'):
                raise Exception(f'Graylog port is not defined for "{kwargs["host"]}"')
            if not kwargs.get('type'):
                raise Exception(f'Graylog type is not defined for "{kwargs["host"]}:{kwargs["port"]}"')
            handler_options: Dict[str, Union[str, bool, int, None]] = {
                'host': kwargs['host'],
                'port': kwargs['port'],
                '_app_name': kwargs.get('appname', 'kubemarine'),
                'debug': kwargs.get('debug', False),
                'version': kwargs.get('version', '1.1')
            }
            if kwargs['type'] == 'tcp':
                self.handler = GelfTcpHandler(**handler_options)
            elif kwargs['type'] == 'udp':
                handler_options['compress'] = kwargs.get('compress', True)
                handler_options['chunk_size'] = kwargs.get('chunk_size', 1300)
                self.handler = GelfUdpHandler(**handler_options)
            elif kwargs['type'] == 'tls':
                handler_options['validate'] = kwargs.get('validate', True)
                handler_options['ca_certs'] = kwargs.get('ca_certs')
                handler_options['certfile'] = kwargs.get('certfile')
                handler_options['keyfile'] = kwargs.get('keyfile')
                self.handler = GelfTlsHandler(**handler_options)
            elif kwargs['type'] == 'http':
                handler_options['compress'] = kwargs.get('compress', True)
                handler_options['path'] = kwargs.get('path', '/gelf')
                handler_options['timeout'] = kwargs.get('timeout', 5)
                self.handler = GelfHttpHandler(**handler_options)
            else:
                raise Exception(f'Unknown Graylog type "{kwargs["type"]}" for "{kwargs["host"]}:{kwargs["port"]}"')

            self.handler.setFormatter(self._formatter)
        else:
            self._target = target
            # Output produced by remote commands might contain characters which cannot be encoded on Windows deployer.
            # Specify explicitly utf-8 encoding which is native to the remote machines.
            self.handler = FileHandlerWithHeader(self._formatter, self._target,
                                                 mode=filemode, header=self._header, encoding='utf-8')

        if level not in LOGGING_LEVELS_BY_NAME:
            raise Exception(f'Failed to create logger - unknown logging level: "{level}"')
        self._level = LOGGING_LEVELS_BY_NAME[level]
        self.handler.setLevel(self._level)

    def __str__(self) -> str:
        return (f'target: {self._target}, level: {LOGGING_NAMES_BY_LEVEL[self._level]}, '
                f'colorize: {self._colorize}, datefmt: {self._datefmt}, format: {self._format}')

    def append_to_logger(self, logger: EnhancedLogger) -> None:
        logger.addHandler(self.handler)

    def has_stdout_target(self) -> bool:
        return self._target == 'stdout'


class Log:

    def __init__(self, name: str, handlers: List[LogHandler]):
        logger = logging.getLogger(name)
        self._logger = cast(EnhancedLogger, logger)
        self._logger.setLevel(VERBOSE)

        if self._logger.hasHandlers():
            self._logger.handlers.clear()

        for handler in handlers:
            handler.append_to_logger(self._logger)

    @property
    def logger(self) -> EnhancedLogger:
        return self._logger


class LoggerWriter:
    stream = sys.stdout.isatty()

    def __init__(self, logger: EnhancedLogger, caller: dict, prefix: str) -> None:
        self.logger = logger
        self.caller = caller
        self.prefix = prefix
        self.buf = ""

    def write(self, message: str) -> None:
        # Both remote stderr and stdout are printed to local stdout.
        # For a non-tty stdout, we should not stream the remote (potentially pty) output immediately.
        # The output should be buffered by lines with converted CRs. See `LoggerWriter._log()`.
        if LoggerWriter.stream:
            sys.stdout.write(message)

        lines = message.split('\n')
        for line in lines[:-1]:
            self.buf = self.buf + line
            self._log()
        self.buf = self.buf + lines[-1]

    def flush(self, remainder: bool = False) -> None:
        if remainder and self.buf:
            self._log()

    def _log(self) -> None:
        line = self.buf
        # Algorithms of CR conversion respects the `connections.RemoteRunner.generate_result()`.
        if not LoggerWriter.stream:
            line = line.rstrip('\r').replace("\r", "\n")
            sys.stdout.write(line + '\n')

        self.logger.log(logging.DEBUG, line, extra={
            'real_caller': self.caller, 'prefix': self.prefix, 'ignore_stdout': True
        })
        self.buf = ""

    def __repr__(self) -> str:
        return f"LoggerWriter{{DEBUG,stdout}} at {hex(id(self))}"


def parse_log_argument(argument: str) -> LogHandler:
    """
    Parse raw CLI arguments and verify for required parameters
    :param argument: Raw CLI argument string. For example: test.log;level=verbose;colorize=true
    :return: Initialized LogHandler
    """
    parameters: Dict[str, Any] = {}
    argument_parts = argument.split(';')
    if not argument_parts:
        raise Exception('Defined logger do not contain parameters')
    target = argument_parts[0]
    level: Optional[str] = None
    for parameter in argument_parts[1:]:
        if parameter == '':
            continue
        value: Union[str, bool, int]
        key, value, *_ = parameter.split('=')
        if key == 'level':
            level = value
        else:
            if key in ['colorize', 'correct_newlines', 'debug', 'compress', 'validate']:
                value = value.lower() in ['true', '1']
            elif key in ['chunk_size', 'timeout', 'port']:
                value = int(value)

            parameters[key] = value

    if level is None:
        raise Exception(f'Logging level is not set for logger "{target}"')
    return LogHandler(target, level, **parameters)


def get_dump_debug_filepath(context: dict) -> Optional[str]:
    args = context['execution_arguments']
    if args['disable_dump']:
        return None

    return os.path.join(args['dump_location'], 'dump', 'debug.log')


def init_log_from_context_args(globals_: dict, context: dict, name: str) -> Log:
    """
    Create Log from raw CLI arguments in Cluster context
    :param globals_: parsed globals collection
    :param context: context holding execution arguments.
    :param name: desirable logger name
    :return: Initialized Log, based on all parsed logging arguments
    """

    handlers = []
    stdout_specified = False

    args = context['execution_arguments']
    if args.get('log') is not None:
        for argument in args.get('log'):
            handler = parse_log_argument(argument[0])
            if handler.has_stdout_target():
                if stdout_specified:
                    raise Exception('Multiple stdout logs specified')
                else:
                    stdout_specified = True
            handlers.append(handler)

    debug_filepath = get_dump_debug_filepath(context)
    if debug_filepath:
        handlers.append(LogHandler(target=debug_filepath,
                                   **globals_['logging']['default_targets']['dump']))

    if not stdout_specified:
        stdout_settings = globals_['logging']['default_targets']['stdout']
        handlers.append(LogHandler(target='stdout', **stdout_settings))

    log = Log(name, handlers)

    log.logger.verbose('Using the following loggers: \n\t%s' % "\n\t".join("- " + str(x) for x in handlers))

    return log


def caller_info(logger: EnhancedLogger) -> Dict[str, object]:
    """
    Catches and returns invocation metadata of the method that calls caller_info()

    :param logger: EnhancedLogger
    :return: dictionary with the invocation metadata
    """
    if sys.version_info[:2] >= (3, 11):
        # https://github.com/python/cpython/pull/28287
        fn, lno, func, sinfo = logger.findCaller(stacklevel=3)
    else:
        fn, lno, func, sinfo = logger.findCaller()

    record: logging.LogRecord = logger.makeRecord("", logging.DEBUG, fn, lno, "", (), None,
                                                  func=func, extra=None, sinfo=sinfo)
    return dict(item for item in record.__dict__.items()
                if item[0] in (
                    # record's fields describing invocation origin
                    'pathname', 'filename', 'module', 'lineno', 'funcName', 'stack_info',
                    # record's fields describing initial process and thread context
                    'thread', 'threadName', 'process', 'processName'
                ))



# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import difflib
import hashlib
import io
import ipaddress
import json
import os
import re
import shutil
import sys
import time
import tarfile
import uuid
from collections import OrderedDict
from copy import deepcopy
from datetime import datetime

from typing import Tuple, Callable, List, TextIO, cast, Union, Dict, Sequence, Optional, NoReturn, BinaryIO

import deepdiff  # type: ignore[import-untyped]
import yaml
import ruamel.yaml

from pathvalidate import sanitize_filepath
from ruamel.yaml import CommentedMap
from useful_types import SupportsAllComparisons

from kubemarine.core import log
from kubemarine.core.errors import pretty_print_error


def do_fail(message: str = '', reason: BaseException = None, hint: str = '', logger: log.EnhancedLogger = None) -> NoReturn:

    if not logger:
        sys.stderr.write("\033[91m")

    pretty_print_error(message, reason, logger)

    # Please do not rewrite this to logging approach:
    # hint should be visible only in stdout and without special formatting
    if hint != "":
        sys.stderr.write("\n")
        sys.stderr.write(hint + "\n")

    if not logger:
        sys.stderr.write("\033[0m")

    sys.exit(1)


def get_elapsed_string(start: float, end: float) -> str:
    elapsed = end - start
    hours, remainder = divmod(elapsed, 3600)
    minutes, seconds = divmod(remainder, 60)
    return '{:02}h {:02}m {:02}s'.format(int(hours), int(minutes), int(seconds))


def prepare_dump_directory(context: dict) -> None:
    args: dict = context['execution_arguments']
    location = args['dump_location']
    reset_directory = not args['disable_dump_cleanup']
    dumpdir = os.path.join(location, 'dump')
    if reset_directory and os.path.exists(dumpdir) and os.path.isdir(dumpdir):
        shutil.rmtree(dumpdir)

    if not args['disable_dump']:
        os.makedirs(dumpdir, exist_ok=True)


def make_ansible_inventory(location: str, c: object) -> None:
    from kubemarine.core.cluster import KubernetesCluster  # pylint: disable=cyclic-import
    cluster = cast(KubernetesCluster, c)

    inventory = cluster.inventory
    roles = []
    for node in inventory['nodes']:
        for role in node['roles']:
            if role not in roles:
                roles.append(role)

    config: dict = {
        'all': [
            'localhost ansible_connection=local'
        ],
        'cluster:children': []
    }

    already_global_defined = []

    for role in roles:
        config[role] = []
        config['cluster:children'].append(role)
        for node in cluster.nodes[role].get_ordered_members_configs_list():
            record = "%s ansible_host=%s ansible_ssh_user=%s ansible_ssh_pass=%s ansible_ssh_private_key_file=%s ip=%s" % \
                     (node['name'],
                      node['connect_to'],
                      node.get('username', cluster.globals['connection']['defaults']['username']),
                      node.get('password'),
                      node.get('keyfile'),
                      node['internal_address'])
            if node.get('address') is not None:
                record += ' external_ip=%s' % node['address']

            if node['name'] not in already_global_defined:
                config['all'].append(record)
                # to avoid duplicate definition in global section we have to check is that was already defined?
                already_global_defined.append(node['name'])

            config[role].append(node['name'])

    config['cluster:vars'] = [
        'ansible_become=True'
    ]

    for group in ['services', 'plugins']:
        for service_name, service_configs in inventory.get(group, {}).items():
            # write to inventory only plugins, which will be installed
            if group == 'plugins' and not service_configs.get('install', False):
                continue

            config['cluster:vars'].append('\n# %s.%s' % (group, service_name))

            if isinstance(service_configs, dict):

                for config_name, config_value in service_configs.items():
                    if config_name in ('installation', 'install'):
                        continue

                    if isinstance(config_value, (dict, list)):
                        config_value = json.dumps(config_value)
                    config['cluster:vars'].append('%s_%s=%s' % (
                        # TODO: Rewrite replace using regex
                        service_name.replace('-', '_').replace('.', '_').replace('/', '_'),
                        config_name.replace('-', '_').replace('.', '_').replace('/', '_'),
                        config_value))
            else:
                config_value = json.dumps(service_configs)
                config['cluster:vars'].append('%s=%s' % (
                    service_name.replace('-', '_').replace('.', '_'),
                    config_value))

    config_compiled = ''
    for section_name, strings in config.items():
        config_compiled += '[%s]' % section_name
        for string in strings:
            config_compiled += '\n' + string
        config_compiled += '\n\n'

    dump_file({}, config_compiled, location, dump_location=False)


def get_current_timestamp_formatted() -> str:
    return datetime.now().strftime("%Y-%m-%d-%H-%M-%S")


def merge_vrrp_ips(procedure_inventory: dict, inventory: dict) -> None:
    # This method is currently unused.
    # If it is ever supported when adding and removing node,
    # it will be necessary to more accurately install and reconfigure the keepalived on existing nodes.
    # Also, it is desirable to change the section format, for example, as for etc_hosts.

    if "vrrp_ips" in inventory and len(inventory["vrrp_ips"]) > 0:
        raise Exception("vrrp_ips section already defined, merging not supported yet")
    else:
        inventory["vrrp_ips"] = procedure_inventory["vrrp_ips"]

    if isinstance(inventory, OrderedDict):
        inventory.move_to_end("vrrp_ips", last=False)


def is_dump_allowed(context: dict, filename: str) -> bool:
    args = context['execution_arguments']
    if args['disable_dump'] \
            and not (filename in ClusterStorage.PRESERVED_DUMP_FILES
                     and context['preserve_inventory']
                     and not args.get('without_act', False)):
        return False

    return True


def dump_file(context: Union[dict, object], data: Union[TextIO, str], filename: str,
              *, dump_location: bool = True) -> None:
    if dump_location:
        if not isinstance(context, dict):
            # cluster is passed instead of the context directly
            from kubemarine.core.cluster import KubernetesCluster  # pylint: disable=cyclic-import
            cluster = cast(KubernetesCluster, context)
            context = cluster.context

        if not is_dump_allowed(context, filename):
            return

        target_path = get_dump_filepath(context, filename)
    else:
        target_path = get_external_resource_path(filename)
    # sanitize_filepath is needed for windows/macOS, where some symbols are restricted in file path,
    # but they can appear in target path. They will be replaced with '_'
    target_path = sanitize_filepath(target_path, replacement_text='_')

    os.makedirs(os.path.dirname(target_path), exist_ok=True)

    if isinstance(data, io.StringIO):
        text = data.getvalue()
    elif isinstance(data, str):
        text = data
    else:
        text = data.read()

    with open_utf8(target_path, 'w') as file:
        file.write(text)


def get_dump_directory(context: dict) -> str:
    parts = [context['execution_arguments']['dump_location'], 'dump']
    if 'dump_subdir' in context:
        parts.append(context['dump_subdir'])

    dump_directory: str = os.path.join(*parts)
    return dump_directory


def get_dump_filepath(context: dict, filename: str) -> str:
    return get_external_resource_path(os.path.join(get_dump_directory(context), filename))


def wait_command_successful(logger: log.EnhancedLogger, attempt: Callable[[], bool],
                            retries: int, timeout: int) -> None:
    while retries > 0:
        logger.debug("Waiting for command to succeed, %s retries left" % retries)
        result = attempt()
        if result:
            logger.debug("Command succeeded")
            return
        retries = retries - 1
        time.sleep(timeout)
    raise Exception("Command failed")


def open_utf8(path: str, mode: str = 'r') -> TextIO:
    return cast(TextIO, open(path, mode + 't', encoding='utf-8'))


def open_internal(path: str, mode: str = 'r') -> TextIO:
    return open_utf8(get_internal_resource_path(path), mode)


def open_external(path: str, mode: str = 'r') -> TextIO:
    return open_utf8(get_external_resource_path(path), mode)


def read_internal(path: str) -> str:
    with open_internal(path) as f:
        return f.read()


def read_external(path: str) -> str:
    with open_external(path) as f:
        return f.read()


def get_external_resource_path(path: str) -> str:
    return os.path.abspath(path)


def get_internal_resource_path(path: str) -> str:
    return os.path.abspath(
        os.path.join(os.path.dirname(__file__), '..', path)
    )


def determine_resource_absolute_file(path: str) -> Tuple[str, bool]:
    """
    Get and verify absolute path to resource file
    :param path: Relative path to resource
    :return: Tuple of absolute path to resource file and flag defining if is an external resource
    """
    # is resource exists as it is defined?
    initial_definition = get_external_resource_path(path)
    if os.path.isfile(initial_definition):
        return initial_definition, True

    # is resource exists as internal resource?
    patched_definition = get_internal_resource_path(path)
    if os.path.isfile(patched_definition):
        return patched_definition, False

    raise Exception('Requested resource %s is not exists at %s or %s' % (path, initial_definition, patched_definition))


def determine_resource_absolute_dir(path: str) -> Tuple[str, bool]:
    """
    Get and verify absolute path to resource directory
    :param path: Relative path to resource
    :return: Tuple of absolute path to resource directory and flag defining if is an external resource
    """
    dirname = os.path.dirname(path)
    # is resource dir exists as it is defined?
    initial_definition = get_external_resource_path(dirname)
    if os.path.isdir(initial_definition):
        return initial_definition, True

    # is resource dir exists as internal resource?
    patched_definition = get_internal_resource_path(dirname)
    if os.path.isdir(patched_definition):
        return patched_definition, False

    raise Exception(
        'Requested resource directory %s is not exists at %s or %s' % (path, initial_definition, patched_definition))


def get_local_file_sha1(filename: str) -> str:
    with open(filename, 'rb') as f:
        return get_stream_sha1(f)


def get_stream_sha1(stream: BinaryIO) -> str:
    sha1 = hashlib.sha1()

    # Read by chunks of 2^16 bytes (65536) and calculate aggregated SHA1
    while True:
        data = stream.read(2 ** 16)
        if not data:
            break
        sha1.update(data)

    return sha1.hexdigest()


def get_remote_tmp_path(filename: str = None, ext: str = None) -> str:
    if filename is None:
        filename = uuid.uuid4().hex

    if ext is not None:
        filename += '.' + ext

    return "/tmp/" + filename


def yaml_structure_preserver() -> ruamel.yaml.YAML:
    """YAML loader and dumper which saves original structure"""
    ruamel_yaml = ruamel.yaml.YAML()
    ruamel_yaml.preserve_quotes = True
    return ruamel_yaml


def deepcopy_yaml(data: dict) -> dict:
    if isinstance(data, CommentedMap):
        ruamel_yaml = yaml_structure_preserver()
        # Dump and parse yaml object to work around ruamel.yaml bug
        # https://sourceforge.net/p/ruamel-yaml/tickets/410/
        buf = io.StringIO()
        ruamel_yaml.dump(data, buf)
        data = ruamel_yaml.load(buf.getvalue())
        return data

    return deepcopy(data)


def subdict_yaml(data: dict, keys: Sequence[str]) -> dict:
    items = (item for item in data.items() if item[0] in keys)
    if isinstance(data, CommentedMap):
        return CommentedMap(items)
    else:
        return dict(items)


def convert_native_yaml(data: dict) -> dict:
    if isinstance(data, CommentedMap):
        buf = io.StringIO()
        yaml_structure_preserver().dump(data, buf)
        data = yaml.safe_load(io.StringIO(buf.getvalue()))

    return data


def identity(x: str) -> str:
    return x


def is_sorted(l: Sequence[str], key: Callable[[str], SupportsAllComparisons] = None) -> bool:
    """
    Check that the specified list is sorted.

    :param l: list to check
    :param key: custom key function to customize the sort order
    :return: boolean flag if the list is sorted
    """
    if key is None:
        key = identity
    return all(key(l[i]) <= key(l[i + 1]) for i in range(len(l) - 1))


def map_sorted(map_: CommentedMap, key: Callable[[str], SupportsAllComparisons] = None) -> CommentedMap:
    """
    Check that the specified CommentedMap is sorted, or create new sorted map from it otherwise.

    :param map_: CommentedMap instance to check
    :param key: custom key function to customize the sort order of the map keys
    :return: the same or new sorted instance of the map
    """
    if key is not None:
        _key = key
    else:
        _key = identity
    map_keys: List[str] = list(map_)
    if not is_sorted(map_keys, key=_key):
        map_ = CommentedMap(sorted(map_.items(), key=lambda item: _key(item[0])))

    return map_


def insert_map_sorted(map_: CommentedMap, k: str, v: object, key: Callable[[str], SupportsAllComparisons] = None) -> None:
    """
    Insert new item to the CommentedMap or update the value for the existing key.
    The map should be already sorted.

    :param map_: sorted CommentedMap instance
    :param k: new key
    :param v: new value
    :param key: custom key function to customize the sort order of the map keys
    """
    if k in map_:
        map_[k] = v
        return

    if key is None:
        key = identity
    # Find position to insert new item maintaining the order
    pos = max((mi + 1 for mi, mv in enumerate(map_)
               if key(mv) < key(k)),
              default=0)

    map_.insert(pos, k, v)


def load_yaml(filepath: str) -> dict:
    try:
        with open_utf8(filepath, 'r') as stream:
            data: dict = yaml.safe_load(stream)
            return data
    except yaml.YAMLError as exc:
        do_fail(f"Failed to load {filepath}", exc)


def pretty_path(path: Sequence[Union[str, int]]) -> str:
    if not path:
        return ""
    return f"[{']['.join(map(repr, path))}]"


def strtobool(value: Union[str, bool]) -> bool:
    """
    The method check string and boolean value
    :param value: Value that should be checked
    """
    val = str(value).lower()
    if val in ('y', 'yes', 't', 'true', 'on', '1'):
        return True
    elif val in ('n', 'no', 'f', 'false', 'off', '0'):
        return False
    else:
        raise ValueError(f"invalid truth value {value!r}")


def strtoint(value: Union[str, int]) -> int:
    if isinstance(value, int):
        return value

    try:
        # whitespace required because python's int() ignores them
        return int(value.replace(' ', '.'))
    except ValueError:
        raise ValueError(f"invalid integer value {value!r}") from None


def print_diff(logger: log.EnhancedLogger, diff: deepdiff.DeepDiff) -> None:
    # Extra transformation to JSON is necessary,
    # because DeepDiff.to_dict() returns custom nested classes that cannot be serialized to yaml by default.
    logger.debug(yaml.safe_dump(yaml.safe_load(diff.to_json())))


def get_unified_diff(old: str, new: str, fromfile: str = '', tofile: str = '') -> Optional[str]:
    diff = list(difflib.unified_diff(
        old.splitlines(), new.splitlines(),
        fromfile=fromfile, tofile=tofile,
        lineterm=''))

    if diff:
        return '\n'.join(diff)

    return None


def get_yaml_diff(old: str, new: str, fromfile: str = '', tofile: str = '') -> Optional[str]:
    if yaml.safe_load(old) == yaml.safe_load(new):
        return None

    return get_unified_diff(old, new, fromfile, tofile)


def isipv(address: str, versions: List[int]) -> bool:
    return ipaddress.ip_network(address).version in versions


def get_version_filepath() -> str:
    return get_internal_resource_path("version")


def get_version() -> str:
    return read_internal(get_version_filepath()).strip()


def minor_version(version: str) -> str:
    """
    Converts vN.N.N to vN.N
    """
    return 'v' + '.'.join(map(str, _test_version(version, 3)[0:2]))


def major_version(version: str) -> str:
    """
    Converts vN.N.N to vN
    """
    return 'v' + '.'.join(map(str, _test_version(version, 3)[0:1]))


def version_key(version: str) -> Tuple[int, int, int]:
    """
    Converts vN.N.N to (N, N, N) that can be used in comparisons.
    """
    v = _test_version(version, 3)
    return v[0], v[1], v[2]


def minor_version_key(version: str) -> Tuple[int, int]:
    """
    Converts vN.N to (N, N) that can be used in comparisons.
    """
    v = _test_version(version, 2)
    return v[0], v[1]


def _test_version(version: str, numbers_amount: int) -> List[int]:
    # catch version without "v" at the first symbol
    is_rc = 0
    if version.startswith('v'):
        version_list: list = version[1:].split('.')
        # catch version with unexpected number or parts
        parts_num = len(version_list)
        try:
            for i, value in enumerate(version_list):
                # catch release candidate version like v1.29.0-rc.1
                if parts_num == 4 and i == 2 and value.endswith('-rc'):
                    value = value[:-3]
                    is_rc = 1
                # whitespace required because python's int() ignores them
                version_list[i] = int(value.replace(' ', '.'))
        except ValueError:
            pass
        else:
            if numbers_amount == parts_num - is_rc:
                return version_list[:numbers_amount]

    expected_pattern = 'v' + '.'.join('N+' for _ in range(numbers_amount))
    if numbers_amount == 3:
        expected_pattern += '[-rc.N+]'
    raise ValueError(f'Incorrect version \"{version}\" format, expected version pattern is \"{expected_pattern}\"')


def parse_aligned_table(table_text: str) -> List[Dict[str, str]]:
    """
    Parse aligned table into the list of {header: cell} map.
    The text with table can be initially produced, for example, by https://pkg.go.dev/text/tabwriter.

    :param table_text: aligned table as string
    :return: List of rows
    """
    rows = table_text.strip().split('\n')

    # Parse headers and their positions. No leading or trailing spaces are expected in the line.
    headers_line = rows[0]
    headers = []
    headers_pos = []
    pos = 0
    for match in re.finditer(r'\s+', headers_line):
        headers_pos.append(pos)
        span = match.span()
        headers.append(headers_line[pos:span[0]])
        pos = span[1]

    headers_pos.append(pos)
    headers.append(headers_line[pos:len(headers_line)])

    # Parse each row in the table,
    # provided that the columns start at the same positions as the headers
    headers_num = len(headers)
    data = []
    for row_text in rows[1:]:
        row = {}
        for i, header in enumerate(headers):
            if i == headers_num - 1:
                cell = row_text[headers_pos[i]:]
            else:
                cell = row_text[headers_pos[i]:headers_pos[i+1]]

            row[header] = cell.strip()

        data.append(row)

    return data


class ClusterStorage:
    """
    File preservation:
    1- Create folder where dumps are stored
    2- Rotating dumps in the storage folder
    3- Uploading dumps to nodes
    4- Copying dumps to new nodes
    """

    PRESERVED_DUMP_FILES = ['procedure.yaml', 'procedure_parameters',
                            'cluster.yaml', 'cluster_initial.yaml', 'cluster_finalized.yaml']

    def __init__(self, cluster: object, context: dict):
        from kubemarine.core.cluster import KubernetesCluster  # pylint: disable=cyclic-import
        self.cluster = cast(KubernetesCluster, cluster)
        self.context = context
        self.dir_path = "/etc/kubemarine/procedures/"
        self.dir_location = ''
        self.local_archive_path = ''

    def make_dir(self) -> None:
        """
        This method creates a directory in which logs about operations on the cluster will be stored.
        """
        readable_timestamp = datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
        initial_procedure = self.context["initial_procedure"]
        dir_name = readable_timestamp + "_" + initial_procedure + "/"
        self.dir_location = self.dir_path + dir_name
        self.cluster.nodes['control-plane'].sudo(f"mkdir -p {self.dir_location} ; sudo rm {self.dir_path + 'latest_dump'} ;"
                                                 f" sudo ln -s {self.dir_location} {self.dir_path + 'latest_dump'}")

    def upload_and_rotate(self) -> None:
        """
        This method uploads and unpacks the archive,
        then packs files with logs and maintains a structured storage of logs on the cluster.
        """
        control_planes = self.cluster.nodes["control-plane"]

        self.cluster.log.debug('Uploading archive with preserved information about the procedure.')
        remote_archive = self.dir_location + "local.tar.gz"
        control_planes.put(self.local_archive_path, remote_archive, sudo=True, compare_hashes=False)
        control_planes.sudo(
            f'tar -C {self.dir_location} -xzv --no-same-owner -f {remote_archive}  && '
            f'sudo rm -f {remote_archive} ')

        not_pack_file = self.cluster.inventory['procedure_history']['archive_threshold']
        delete_old = self.cluster.inventory['procedure_history']['delete_threshold']

        command = f'ls {self.dir_path} | grep -v latest_dump'
        node_group_results = control_planes.sudo(command)
        with control_planes.new_executor() as exe:
            for control_plane in exe.group.get_ordered_members_list():
                result = node_group_results[control_plane.get_host()]
                files = result.stdout.split()
                files.sort(reverse=True)
                for i, file in enumerate(files):
                    if not_pack_file <= i < delete_old:
                        if 'tar.gz' not in file:
                            control_plane.sudo(
                                f'tar -czvf {self.dir_path + file + ".tar.gz"} {self.dir_path + file} &&'
                                f'sudo rm -r {self.dir_path + file}')
                    elif i >= delete_old:
                        control_plane.sudo(f'rm -rf {self.dir_path + file}')

    def compress_archive(self, enriched: bool) -> None:
        """
        This method compose dump files in the local archive.
        """
        context = self.context
        self.local_archive_path = get_dump_filepath(context, "local.tar.gz")
        with tarfile.open(self.local_archive_path, "w:gz") as tar:
            dump_files = set(ClusterStorage.PRESERVED_DUMP_FILES)
            if not enriched:
                dump_files -= {'cluster.yaml', 'cluster_finalized.yaml'}

            for name in dump_files:
                source = get_dump_filepath(context, name)
                if os.path.exists(source):
                    tar.add(source, 'dump/' + name)
            tar.add(context['execution_arguments']['config'], 'cluster.yaml')
            tar.add(get_version_filepath(), 'version')

    def collect_procedure_info(self) -> None:
        """
        This method collects information about the type of procedure and the version of the tool we are working with.
        """
        context = self.context
        out = {}
        out['arguments'] = context['initial_cli_arguments']
        if 'proceeded_tasks' in context:
            out['finished_tasks'] = context['proceeded_tasks']
        out["initial_procedure"] = context["initial_procedure"]
        out["successfully_performed"] = context["successfully_performed"]
        out['status'] = context['status']
        output = yaml.dump(out)
        dump_file(context, output, "procedure_parameters")

    def upload_info_new_control_planes(self) -> None:
        """
        This method is used to transfer backup logs from the initial control-plane to the new control-planes.
        """
        new_control_planes = self.cluster.get_new_nodes().having_roles(['control-plane'])
        if new_control_planes.is_empty():
            return

        archive_name = 'dump_log_cluster.tar.gz'
        archive_dump_path = get_dump_filepath(self.context, archive_name)
        archive_remote_path = f"/tmp/{archive_name}"
        log = self.cluster.log

        control_plane = self.cluster.previous_nodes['control-plane'].get_first_member()
        data_copy_res = control_plane.sudo(f'tar -czvf {archive_remote_path} {self.dir_path}')
        log.verbose("Archive with procedures history is created:\n%s" % data_copy_res)
        control_plane.get(archive_remote_path, archive_dump_path)

        log.debug("Archive with procedures history is downloaded")

        for group in new_control_planes.get_ordered_members_list():
            group.put(archive_dump_path, archive_remote_path, sudo=True, compare_hashes=False)
            group.sudo(f'tar -C / -xzvf {archive_remote_path}')
            log.debug(f"Archive with procedures history is uploaded to {group.get_node_name()!r}")



#!/usr/bin/env python3
# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from collections import OrderedDict
from typing import List

from kubemarine import plugins, k8s_certs
from kubemarine.core import flow
from kubemarine.core.cluster import KubernetesCluster
from kubemarine.plugins import calico


def renew_nginx_ingress_certs_task(cluster: KubernetesCluster) -> None:
    # check that renewal is required for nginx
    if not cluster.procedure_inventory.get("nginx-ingress-controller"):
        cluster.log.debug("Skipped: nginx ingress controller certs renewal is not required")
        return

    cluster.log.debug("Starting certificate renewal for nginx ingress controller, plugin will be reinstalled")
    plugin = cluster.inventory["plugins"]["nginx-ingress-controller"]
    # If by chance default certificate is not previously configured,
    # the procedure reconfigures the DeamonSet as well.
    plugins.install_plugin(cluster, "nginx-ingress-controller", plugin["installation"]['procedures'])


def renew_calico_apiserver_certs_task(cluster: KubernetesCluster) -> None:
    # check that renewal is required for the Calico API server
    if 'calico' not in cluster.procedure_inventory:
        cluster.log.debug("Skipped: Calico API server certs renewal is not required")
        return

    # Let's assume that if the user specified `calico` section in the procedure inventory,
    # they agree with the default renew procedure.
    # Also, it implies that the `calico` plugin is enabled and installed, and Calico API server is enabled.
    calico.renew_apiserver_certificate(cluster)


def k8s_certs_renew_task(cluster: KubernetesCluster) -> None:
    if not cluster.procedure_inventory.get("kubernetes"):
        cluster.log.debug("Skipped: kubernetes certs renewal is not required")
        return

    cluster.log.debug("Starting certificate renewal for kubernetes")
    cluster.nodes['control-plane'].call(k8s_certs.renew_apply)


def k8s_certs_overview_task(cluster: KubernetesCluster) -> None:
    cluster.nodes['control-plane'].call(k8s_certs.k8s_certs_overview)


tasks = OrderedDict({
    "kubernetes": k8s_certs_renew_task,
    "nginx_ingress_controller": renew_nginx_ingress_certs_task,
    "calico": renew_calico_apiserver_certs_task,
    "certs_overview": k8s_certs_overview_task
})


class CertRenewAction(flow.TasksAction):
    def __init__(self) -> None:
        super().__init__('cert renew', tasks, recreate_inventory=True)


def create_context(cli_arguments: List[str] = None) -> dict:

    cli_help = '''
    Script for certificates renewal on existing Kubernetes cluster.

    How to use:

    '''

    parser = flow.new_procedure_parser(cli_help, tasks=tasks)
    context = flow.create_context(parser, cli_arguments, procedure='cert_renew')
    return context


def main(cli_arguments: List[str] = None) -> None:
    context = create_context(cli_arguments)
    flow.ActionsFlow([CertRenewAction()]).run_flow(context)


if __name__ == '__main__':
    main()



# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import argparse
import re
import sys
from abc import ABC, abstractmethod
from textwrap import dedent
from typing import List, Union

import yaml

import kubemarine.patches
from kubemarine import kubernetes, plugins, packages, etcd, thirdparties, haproxy, keepalived
from kubemarine.core import flow, static, utils, errors
from kubemarine.core.action import Action
from kubemarine.core.cluster import KubernetesCluster, EnrichmentStage
from kubemarine.core.group import NodeGroup
from kubemarine.core.patch import Patch, _SoftwareUpgradePatch
from kubemarine.core.resources import DynamicResources

SOFTWARE_UPGRADE_PATH = utils.get_internal_resource_path("patches/software_upgrade.yaml")


class SoftwareUpgradeAction(Action, ABC):
    def __init__(self, software_name: str, k8s_versions: List[str]) -> None:
        super().__init__(f'Upgrade {software_name}')
        self.software_name = software_name
        self.k8s_versions = k8s_versions

    def run(self, res: DynamicResources) -> None:
        # Use DEFAULT enrichment to compile inventory, but not apply procedure inventory.
        cluster = res.cluster(EnrichmentStage.DEFAULT)
        if not self.is_patch_relevant(cluster):
            return

        self.recreate_inventory = True

        self._prepare_upgrade_context(res.context)
        res.reset_cluster(EnrichmentStage.DEFAULT)

        # Run full enrichment to apply procedure inventory based on new context.
        cluster = res.cluster(EnrichmentStage.PROCEDURE)
        self._upgrade(cluster)
        # Change context back, but do not DynamicResources.reset_cluster().
        # Once the inventory is recreated, the reset context will be picked up during new cluster initialization.
        self._reset_upgrade_context(res.context)

    def is_patch_relevant(self, cluster: KubernetesCluster) -> bool:
        version = kubernetes.get_kubernetes_version(cluster.inventory)
        if version not in self.k8s_versions:
            cluster.log.info(f"Patch is not relevant for Kubernetes {version}")
            return False

        return True

    @abstractmethod
    def _prepare_upgrade_context(self, context: dict) -> None:
        pass

    @abstractmethod
    def _upgrade(self, cluster: KubernetesCluster) -> None:
        pass

    @abstractmethod
    def _reset_upgrade_context(self, context: dict) -> None:
        return


class ThirdpartyUpgradeAction(SoftwareUpgradeAction):
    def __init__(self, software_name: str, k8s_versions: List[str]) -> None:
        super().__init__(software_name, k8s_versions)
        self._destination = thirdparties.get_thirdparty_destination(self.software_name)

    def _upgrade(self, cluster: KubernetesCluster) -> None:
        cluster.log.info(f"Reinstalling third-party {self._destination!r}")
        self._run(cluster)

    def _run(self, cluster: KubernetesCluster) -> None:
        thirdparties.install_thirdparty(cluster.nodes['all'], self._destination)

    def _prepare_upgrade_context(self, context: dict) -> None:
        context['upgrading_thirdparty'] = self._destination

    def _reset_upgrade_context(self, context: dict) -> None:
        del context['upgrading_thirdparty']


class CriUpgradeAction(Action):
    def __init__(self, upgrade_config: dict) -> None:
        super().__init__(f'Upgrade CRI')
        self.upgrade_config = upgrade_config

    def run(self, res: DynamicResources) -> None:
        # Use DEFAULT enrichment to compile inventory, but not apply procedure inventory.
        cluster = res.cluster(EnrichmentStage.DEFAULT)

        if not self._associations_changed(cluster):
            return

        res.context['upgrading_package'] = 'containerd'
        res.reset_cluster(EnrichmentStage.DEFAULT)

        # Run full enrichment to apply procedure inventory based on new context.
        cluster = res.cluster(EnrichmentStage.PROCEDURE)
        if 'containerd' not in cluster.context["upgrade"]["required"]['packages']:
            res.logger().info(f"Nothing has changed in associations of 'containerd'. Upgrade is not required.")
            self._reset_upgrade_context(res.context)
            res.reset_cluster(EnrichmentStage.DEFAULT)
            return

        self.recreate_inventory = True
        self._run(cluster)
        # Change context back, but do not DynamicResources.reset_cluster().
        # Once the inventory is recreated, the reset context will be picked up during new cluster initialization.
        self._reset_upgrade_context(res.context)

    def _reset_upgrade_context(self, context: dict) -> None:
        del context['upgrading_package']

    def _run(self, cluster: KubernetesCluster) -> None:
        if 'worker' in cluster.nodes:
            self.upgrade_cri(cluster.nodes["worker"].exclude_group(cluster.nodes["control-plane"]), workers=True)
        self.upgrade_cri(cluster.nodes["control-plane"], workers=False)

    def upgrade_cri(self, group: NodeGroup, workers: bool) -> None:
        cluster: KubernetesCluster = group.cluster

        drain_timeout = cluster.procedure_inventory.get('drain_timeout')
        grace_period = cluster.procedure_inventory.get('grace_period')
        disable_eviction = cluster.procedure_inventory.get("disable-eviction", True)

        for node in group.get_ordered_members_list():
            node_name = node.get_node_name()
            control_plane = node
            if workers:
                control_plane = cluster.nodes["control-plane"].get_first_member()

            drain_cmd = kubernetes.prepare_drain_command(
                cluster, node_name,
                disable_eviction=disable_eviction, drain_timeout=drain_timeout, grace_period=grace_period)
            control_plane.sudo(drain_cmd, hide=False, pty=True)

            kubernetes.upgrade_cri_if_required(node)
            node.sudo('systemctl restart kubelet')

            if workers:
                control_plane.sudo(f"kubectl uncordon {node_name}", hide=False)
            else:
                kubernetes.wait_uncordon(node)

            kubernetes.components.wait_for_pods(node)
            if not workers:
                etcd.wait_for_health(cluster, node)

    def _associations_changed(self, cluster: KubernetesCluster) -> bool:
        """
        Detects if upgrade is required for the given Kubernetes version, OS family and CRI implementation.
        """
        version = kubernetes.get_kubernetes_version(cluster.inventory)

        os_family = cluster.get_os_family()
        if os_family not in packages.get_associations_os_family_keys():
            raise errors.KME("KME0012", procedure='migrate_kubemarine')
        version_key = packages.get_compatibility_version_key(os_family)

        changes_detected = False
        packages_names = static.GLOBALS['packages'][os_family]['containerd']['package_name']
        for kv in packages_names:
            software_name = list(kv.values())[0]
            kubernetes_upgrade_list = self.upgrade_config['packages'][software_name][version_key]
            changes_detected = changes_detected or version in kubernetes_upgrade_list

        if not changes_detected:
            cluster.log.info(f"Patch is not relevant for Kubernetes {version}, "
                             f"based on 'containerd' and {os_family!r} OS family")

        return changes_detected


class BalancerUpgradeAction(Action):
    def __init__(self, upgrade_config: dict, package_name: str) -> None:
        super().__init__(f'Upgrade {package_name}')
        self.upgrade_config = upgrade_config
        self.package_name = package_name

    def run(self, res: DynamicResources) -> None:
        # Use DEFAULT enrichment to compile inventory, but not apply procedure inventory.
        cluster = res.cluster(EnrichmentStage.DEFAULT)
        logger = res.logger()

        if not self._associations_changed(cluster):
            return

        role = 'balancer' if self.package_name == 'haproxy' else 'keepalived'
        group = cluster.make_group_from_roles([role])
        if group.is_empty():
            logger.info(f"No nodes to install {self.package_name!r}.")
            return

        self._prepare_upgrade_context(res.context)
        res.reset_cluster(EnrichmentStage.DEFAULT)

        # Run full enrichment to apply procedure inventory based on new context.
        cluster = res.cluster(EnrichmentStage.PROCEDURE)
        if self.package_name not in cluster.context["upgrade"]["required"]['packages']:
            logger.info(f"Nothing has changed in associations of {self.package_name!r}. Upgrade is not required.")
            self._reset_upgrade_context(res.context)
            res.reset_cluster(EnrichmentStage.DEFAULT)
            return

        logger.info(f"Reinstalling {self.package_name} on nodes: {group.get_nodes_names()}")
        self.recreate_inventory = True

        self._run(group)
        # Change context back, but do not DynamicResources.reset_cluster().
        # Once the inventory is recreated, the reset context will be picked up during new cluster initialization.
        self._reset_upgrade_context(res.context)

    def _run(self, group: NodeGroup) -> None:
        cluster: KubernetesCluster = group.cluster
        pkgs = cluster.get_package_association(self.package_name, 'package_name')
        packages.install(group, include=pkgs, pty=True)
        if self.package_name == 'haproxy':
            haproxy.restart(group)
        else:
            keepalived.restart(group)

    def _associations_changed(self, cluster: KubernetesCluster) -> bool:
        """
        Detects if upgrade is required for the given OS family.
        """
        os_family = cluster.get_os_family()
        if os_family not in packages.get_associations_os_family_keys():
            raise errors.KME("KME0012", procedure='migrate_kubemarine')
        version_key = packages.get_compatibility_version_key(os_family)

        changes_detected: Union[List[str], bool] = self.upgrade_config['packages'][self.package_name][version_key]
        if not changes_detected:
            cluster.log.info(f"Patch is not relevant for {os_family!r} OS family")

        return bool(changes_detected)

    def _prepare_upgrade_context(self, context: dict) -> None:
        context['upgrading_package'] = self.package_name

    def _reset_upgrade_context(self, context: dict) -> None:
        del context['upgrading_package']


class PluginUpgradeAction(SoftwareUpgradeAction):
    def _upgrade(self, cluster: KubernetesCluster) -> None:
        # Recommended version is changed in globals.
        # In general, there can be arbitrary references to globals using jinja templates in the inventory,
        # So let's always run upgrade.
        upgrade_candidates = {
            self.software_name: cluster.inventory['plugins'][self.software_name]
        }
        self._run(cluster, upgrade_candidates)

    def _run(self, cluster: KubernetesCluster, upgrade_candidates: dict) -> None:
        plugins.install(cluster, upgrade_candidates)

    def _prepare_upgrade_context(self, context: dict) -> None:
        context['upgrading_plugin'] = self.software_name
        if self.software_name == 'calico':
            context['upgrading_thirdparty'] = thirdparties.get_thirdparty_destination('calicoctl')

    def _reset_upgrade_context(self, context: dict) -> None:
        del context['upgrading_plugin']
        if self.software_name == 'calico':
            del context['upgrading_thirdparty']


class ThirdpartyUpgradePatch(_SoftwareUpgradePatch):
    def __init__(self, thirdparty_name: str, k8s_versions: List[str]) -> None:
        super().__init__(f"upgrade_{thirdparty_name}")
        self.thirdparty_name = thirdparty_name
        self.k8s_versions = k8s_versions

    @property
    def action(self) -> Action:
        return ThirdpartyUpgradeAction(self.thirdparty_name, self.k8s_versions)

    @property
    def description(self) -> str:
        versions_list = '\n'.join(f" - {ver}" for ver in self.k8s_versions)

        return dedent(
            f"""\
            Upgrade {self.thirdparty_name!r} for the following Kubernetes versions:
            {{versions_list}}
            Roughly equivalent to 'kubemarine install --tasks=prepare.thirdparties'
            provided that all third-parties except the {self.thirdparty_name!r} are already actual.
            """.rstrip()
        ).format(versions_list=versions_list)


class CriUpgradePatch(_SoftwareUpgradePatch):
    def __init__(self, upgrade_config: dict) -> None:
        super().__init__(f"upgrade_cri")
        self.upgrade_config = upgrade_config

    @property
    def action(self) -> Action:
        return CriUpgradeAction(self.upgrade_config)

    @property
    def description(self) -> str:
        return dedent(
            f"""\
            Upgrade CRI for necessary Kubernetes versions with particular CRI and particular OS family.
            Exact configuration for what versions the upgrade is necessary, can be seen in
            {SOFTWARE_UPGRADE_PATH}.
            Upgrade procedure is similar to 'kubemarine upgrade --tasks=kubernetes', but without the Kubernetes upgrade.
            """.rstrip()
        )


class BalancerUpgradePatch(_SoftwareUpgradePatch):
    def __init__(self, upgrade_config: dict, package_name: str) -> None:
        super().__init__(f"upgrade_{package_name}")
        self.upgrade_config = upgrade_config
        self.package_name = package_name

    @property
    def action(self) -> Action:
        return BalancerUpgradeAction(self.upgrade_config, self.package_name)

    @property
    def description(self) -> str:
        return dedent(
            f"""\
            Upgrade {self.package_name!r} on balancers:
            Roughly equivalent to 'kubemarine install --tasks=deploy.loadbalancer.{self.package_name}.install',
            as if it is run in a clean environment.
            """.rstrip()
        )


class PluginUpgradePatch(_SoftwareUpgradePatch):
    def __init__(self, plugin_name: str, k8s_versions: List[str]) -> None:
        super().__init__(f"upgrade_{re.sub(r'-', '_', plugin_name)}")
        self.plugin_name = plugin_name
        self.k8s_versions = k8s_versions

    @property
    def action(self) -> Action:
        return PluginUpgradeAction(self.plugin_name, self.k8s_versions)

    @property
    def description(self) -> str:
        versions_list = '\n'.join(f" - {ver}" for ver in self.k8s_versions)

        return dedent(
            f"""\
            Upgrade {self.plugin_name!r} for the following Kubernetes versions:
            {{versions_list}}
            Roughly equivalent to 'kubemarine install --tasks=deploy.plugins'
            with all plugins disabled except the {self.plugin_name!r}.
            """.rstrip()
        ).format(versions_list=versions_list)


def load_upgrade_config() -> dict:
    with utils.open_internal(SOFTWARE_UPGRADE_PATH) as stream:
        upgrade_config: dict = yaml.safe_load(stream)
        return upgrade_config


def resolve_upgrade_patches() -> List[_SoftwareUpgradePatch]:
    upgrade_config = load_upgrade_config()

    upgrade_patches: List[_SoftwareUpgradePatch] = []

    # The order of upgrade is determined by the implementation below

    for thirdparty_name in ['crictl']:
        k8s_versions = upgrade_config['thirdparties'][thirdparty_name]
        if k8s_versions:
            verify_allowed_kubernetes_versions(k8s_versions)
            upgrade_patches.append(ThirdpartyUpgradePatch(thirdparty_name, k8s_versions))

    k8s_versions = [version
                    for pkg in ('containerd', 'containerdio')
                    for v_key in ('version_rhel', 'version_rhel8', 'version_rhel9', 'version_debian')
                    for version in upgrade_config['packages'][pkg].get(v_key, [])]
    if k8s_versions:
        verify_allowed_kubernetes_versions(k8s_versions)
        upgrade_patches.append(CriUpgradePatch(upgrade_config))

    for package_name in ['haproxy', 'keepalived']:
        if any(upgrade_config['packages'][package_name].get(v_key)
               for v_key in ('version_rhel', 'version_rhel8', 'version_rhel9', 'version_debian')):
            upgrade_patches.append(BalancerUpgradePatch(upgrade_config, package_name))

    default_plugins = static.DEFAULTS['plugins']
    plugins = list(default_plugins)
    plugins.sort(key=get_default_plugin_priority)
    for plugin_name in plugins:
        k8s_versions = upgrade_config['plugins'][plugin_name]
        if k8s_versions:
            verify_allowed_kubernetes_versions(k8s_versions)
            upgrade_patches.append(PluginUpgradePatch(plugin_name, k8s_versions))

    return upgrade_patches


def get_default_plugin_priority(plugin: str) -> int:
    priority: int = static.DEFAULTS['plugins'][plugin]['installation']['priority']
    return priority


def verify_allowed_kubernetes_versions(kubernetes_versions: List[str]) -> None:
    not_allowed_versions = set(kubernetes_versions) - set(static.KUBERNETES_VERSIONS['compatibility_map'])
    if not_allowed_versions:
        raise Exception(f"Kubernetes versions {', '.join(map(repr, not_allowed_versions))} are not allowed.")


def load_patches() -> List[Patch]:
    patches = list(kubemarine.patches.patches)
    patches.extend(resolve_upgrade_patches())
    patches.sort(key=lambda p: p.priority())
    return patches


def new_parser() -> argparse.ArgumentParser:
    cli_help = '''
    Script for automated update of the environment for the current version of Kubemarine.

    How to use:

    '''

    parser = flow.new_common_parser(cli_help)
    parser.add_argument('--force-skip', dest='skip', metavar='SKIP',
                        default='',
                        help='define comma-separated patches to skip')

    parser.add_argument('--force-apply', dest='apply', metavar='APPLY',
                        default='',
                        help='define explicit comma-separated set of patches to apply')

    parser.add_argument('--list',
                        action='store_true',
                        help='list all patches')

    parser.add_argument('--describe', metavar='PATCH',
                        help='describe the specified patch')

    parser.add_argument('procedure_config', metavar='procedure_config',
                        type=str, help='config file for the procedure', nargs='?')

    return parser


def run(context: dict) -> None:
    # pylint: disable=bad-builtin

    args = context['execution_arguments']

    patches = load_patches()
    patch_ids = [patch.identifier for patch in patches]

    if args['list']:
        if patch_ids:
            print("Available patches list:")
            for patch_id in patch_ids:
                print(patch_id)
        else:
            print("No patches available.")
        sys.exit(0)

    if args['describe']:
        for patch in patches:
            if patch.identifier == args['describe']:
                print(patch.description)
                sys.exit(0)
        print(f"Unknown patch '{args['describe']}'")
        sys.exit(1)

    skip = [] if not args['skip'] else args['skip'].split(",")
    apply = [] if not args['apply'] else args['apply'].split(",")

    if apply and (set(apply) - set(patch_ids)):
        print(f"Unknown patches {list(set(apply) - set(patch_ids))}")
        sys.exit(1)

    if skip and (set(skip) - set(patch_ids)):
        print(f"Unknown patches {list(set(skip) - set(patch_ids))}")
        sys.exit(1)

    if apply:
        positions = [patch_ids.index(apply_id) for apply_id in apply]
        if not all(positions[i] < positions[i + 1] for i in range(len(positions) - 1)):
            print("Incorrect order of patches to apply. See --list for correct order of patches.")
            sys.exit(1)

    actions = []
    for patch in patches:
        if apply:
            if patch.identifier in apply:
                actions.append(patch.action)
        elif skip:
            if patch.identifier not in skip:
                actions.append(patch.action)
        else:
            actions.append(patch.action)

    if not actions:
        print("No patches to apply")
        sys.exit(0)

    flow.ActionsFlow(actions).run_flow(context)


def create_context(cli_arguments: List[str] = None) -> dict:
    parser = new_parser()
    context = flow.create_context(parser, cli_arguments, procedure="migrate_kubemarine")
    return context


def main(cli_arguments: List[str] = None) -> None:
    context = create_context(cli_arguments)
    run(context)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import concurrent
import datetime
import gzip
import json
import os
import shutil
import tarfile
import threading
import time
import uuid
from collections import OrderedDict
from concurrent.futures import ThreadPoolExecutor
from queue import Queue, Empty
from typing import List, Tuple, Union, Dict, Optional, Iterator, Literal

import yaml

from kubemarine import etcd
from kubemarine.core import utils, flow, log, connections
from kubemarine.core.cluster import KubernetesCluster
from kubemarine.core.group import NodeGroup, RemoteExecutor
from kubemarine.cri import containerd


def get_default_backup_files_list(cluster: KubernetesCluster) -> List[str]:
    haproxy_service = cluster.get_package_association('haproxy', 'service_name')
    keepalived_service = cluster.get_package_association('keepalived', 'service_name')

    backup_files_list = [
        "/etc/resolv.conf",
        "/etc/hosts",
        "/etc/chrony.conf",
        "/etc/selinux/config",
        "/etc/yum.repos.d/",
        "/etc/apt/sources.list.d/",
        "/var/lib/kubelet/pki",
        "/etc/modules-load.d/",
        "/etc/audit/rules.d/",
        "/etc/haproxy/haproxy.cfg",
        f"/etc/systemd/system/{haproxy_service}.service.d/{haproxy_service}.conf",
        "/etc/keepalived/keepalived.conf",
        f"/etc/systemd/system/{keepalived_service}.service.d/{keepalived_service}.conf",
        "/usr/local/bin/check_haproxy.sh",
        "/etc/kubernetes",
        "/root/.kube/config",
        "/etc/systemd/system/kubelet.service"
    ]

    backup_files_list.append("/etc/containerd/config.toml")
    backup_files_list.append("/etc/crictl.yaml")
    backup_files_list.append("/etc/ctr/kubemarine_ctr_flags.conf")
    config_path = containerd.get_config_path(cluster.inventory)
    if config_path:
        backup_files_list.append(config_path)


    return backup_files_list


def prepare_backup_tmpdir(logger: log.EnhancedLogger, context: dict) -> str:
    backup_directory = context.get('backup_tmpdir')
    if not backup_directory:
        logger.verbose('Backup directory is not ready yet, preparing..')
        backup_directory = context['backup_tmpdir'] = utils.get_dump_filepath(context, 'backup')
        shutil.rmtree(backup_directory, ignore_errors=True)
        os.mkdir(backup_directory)
        logger.verbose('Backup directory prepared')
    return backup_directory


def verify_backup_location(cluster: KubernetesCluster) -> None:
    target = utils.get_external_resource_path(cluster.procedure_inventory.get('backup_location', 'backup.tar.gz'))
    if not os.path.isdir(target) and not os.path.isdir(os.path.abspath(os.path.join(target, os.pardir))):
        raise FileNotFoundError('Backup location directory not exists')


def export_ansible_inventory(cluster: KubernetesCluster) -> None:
    backup_directory = prepare_backup_tmpdir(cluster.log, cluster.context)
    shutil.copyfile(cluster.context['execution_arguments']['ansible_inventory_location'],
                    os.path.join(backup_directory, 'ansible-inventory.ini'))
    cluster.log.verbose('ansible-inventory.ini exported to backup')


def export_packages_list(cluster: KubernetesCluster) -> None:
    cluster.context['backup_descriptor']['nodes']['packages'] = {}
    if cluster.get_os_family() in ['rhel', 'rhel8', 'rhel9']:
        cmd = r"rpm -qa"
    else:
        cmd = r"dpkg-query -f '${Package}=${Version}\n' -W"
    results = cluster.nodes['all'].sudo(cmd)
    for host, result in results.items():
        cluster.context['backup_descriptor']['nodes']['packages'][host] = result.stdout.strip().split('\n')


def export_hostname(cluster: KubernetesCluster) -> None:
    cluster.context['backup_descriptor']['nodes']['hostnames'] = {}
    results = cluster.nodes['all'].sudo('hostnamectl status | head -n 1 | sed -e \'s/[a-zA-Z ]*://g\'')
    cluster.log.verbose(results)
    for host, result in results.items():
        cluster.context['backup_descriptor']['nodes']['hostnames'][host] = result.stdout.strip()


def export_cluster_yaml(cluster: KubernetesCluster) -> None:
    backup_directory = prepare_backup_tmpdir(cluster.log, cluster.context)
    shutil.copyfile(utils.get_dump_filepath(cluster.context, 'cluster.yaml'),
                    os.path.join(backup_directory, 'cluster.yaml'))
    shutil.copyfile(utils.get_external_resource_path(cluster.context['execution_arguments']['config']),
                    os.path.join(backup_directory, 'original_cluster.yaml'))
    cluster.log.verbose('cluster.yaml exported to backup')


def export_nodes(cluster: KubernetesCluster) -> None:
    backup_directory = prepare_backup_tmpdir(cluster.log, cluster.context)
    backup_nodes_data_dir = os.path.join(backup_directory, 'nodes_data')
    os.mkdir(backup_nodes_data_dir)

    backup_list = get_default_backup_files_list(cluster)
    for filepath, enabled in cluster.procedure_inventory.get('backup_plan', {}).get('nodes', {}).items():
        if not enabled and filepath in backup_list:
            backup_list.remove(filepath)
        if enabled and filepath not in backup_list:
            backup_list.append(filepath)

    cluster.log.debug('Backing up the following files: \n' + '  - ' + '\n  - '.join(backup_list))

    backup_command = 'cd /tmp && ' \
                     'sudo tar -czvf /tmp/kubemarine-backup.tar.gz -P $(sudo readlink -e %s) && ' \
                     'sudo ls -la /tmp/kubemarine-backup.tar.gz && ' \
                     'sudo du -hs /tmp/kubemarine-backup.tar.gz' % (' '.join(backup_list))

    data_copy_res = cluster.nodes['all'].run(backup_command, pty=True)

    cluster.log.debug('Backup created:\n%s' % data_copy_res)

    cluster.log.debug('Downloading nodes backups:')
    for node in cluster.nodes['all'].get_ordered_members_list():
        node.get('/tmp/kubemarine-backup.tar.gz',
                 os.path.join(backup_nodes_data_dir, '%s.tar.gz' % node.get_node_name()))
        cluster.log.debug('Backup \'%s\' downloaded' % node.get_node_name())

    cluster.log.verbose('Deleting backup file from nodes...')
    cluster.nodes['all'].sudo('rm -f /tmp/kubemarine-backup.tar.gz')


def export_etcd(cluster: KubernetesCluster) -> None:
    backup_directory = prepare_backup_tmpdir(cluster.log, cluster.context)
    etcd_node, is_custom_etcd_node = select_etcd_node(cluster)
    cluster.context['backup_descriptor']['etcd']['image'] = retrieve_etcd_image(etcd_node)

    # Try to detect cluster health and other metadata like db size, leader
    etcd_status = None
    try:
        _, etcd_status = etcd.execute_endpoints_command(cluster, etcd_node, 'status', warn=False)
        parsed_etcd_status = {}
        for item in etcd_status:
            # get rid of https:// and :2379
            address = item['endpoint'].split('//')[1].split(':')[0]
            node_name = cluster.nodes['all'].get_first_member(apply_filter={"internal_address": address}).get_node_name()
            parsed_etcd_status[node_name] = item
        cluster.context['backup_descriptor']['etcd']['status'] = parsed_etcd_status
    except Exception:
        cluster.log.verbose('Failed to load and parse ETCD status')

    if is_custom_etcd_node:
        cluster.context['backup_descriptor']['etcd']['source'] = etcd_node.get_node_name()
    else:
        # if user did not provide node, then we have to select leader by ourselves
        if not etcd_status:
            raise Exception('Failed to load ETCD status and impossible to detect ETCD leader for making snapshot from it')
        etcd_leader_id = etcd_status[0]['status']['leader']
        etcd_leader_name = None
        for name, item in parsed_etcd_status.items():
            if item['status']['header']['member_id'] == etcd_leader_id:
                etcd_leader_name = name
        if etcd_leader_name:
            cluster.log.verbose('Detected ETCD leader: %s' % etcd_leader_name)
            cluster.context['backup_descriptor']['etcd']['source'] = etcd_leader_name
            etcd_node = cluster.nodes['control-plane'].get_member_by_name(etcd_leader_name)
        else:
            raise Exception('Failed to detect ETCD leader - not possible to create backup from actual DB')

    snap_name = 'snapshot%s.db' % int(round(time.time() * 1000))
    endpoint_ip = etcd_node.get_config()["internal_address"]
    cluster.log.debug('Creating ETCD backup "%s"...' % snap_name)
    result = etcd_node.sudo(f'etcdctl snapshot save /var/lib/etcd/{snap_name} --endpoints=https://{endpoint_ip}:2379 '
                            f'&& sudo mv /var/lib/etcd/{snap_name} /tmp/{snap_name} '
                            f'&& sudo ls -la /tmp/{snap_name} '
                            f'&& sudo du -hs /tmp/{snap_name} '
                            f'&& sudo chmod 666 /tmp/{snap_name}',
                            pty=True, timeout=600)
    cluster.log.debug(result)
    etcd_node.get('/tmp/' + snap_name, backup_directory + '/etcd.db')
    cluster.log.verbose('Deleting ETCD snapshot file from "%s"...')
    etcd_node.sudo('rm -f /tmp/%s' % snap_name)


def select_etcd_node(cluster: KubernetesCluster) -> Tuple[NodeGroup, bool]:
    custom_etcd_node = cluster.procedure_inventory.get('backup_plan', {}).get('etcd', {}).get('source_node')

    if custom_etcd_node:
        if not cluster.nodes['all'].has_node(custom_etcd_node):
            raise Exception('Unknown ETCD node selected as source')
        etcd_node = cluster.nodes['all'].get_member_by_name(custom_etcd_node)
        return etcd_node, True
    else:
        return cluster.nodes['control-plane'].get_any_member(), False


def retrieve_etcd_image(etcd_node: NodeGroup) -> str:
    # TODO: Detect ETCD version via /etc/kubernetes/manifests/etcd.yaml config if presented, otherwise use containers
    node_name = etcd_node.get_node_name()

    cont_search = "sudo crictl ps --label io.kubernetes.pod.name=etcd-%s -aq | awk '{print $1; exit}'" % node_name
    cont_inspect = f"crictl inspect $({cont_search})"
    etcd_container_json = json.loads(list(etcd_node.sudo(cont_inspect).values())[0].stdout)
    etcd_image_sha = etcd_container_json['info']['config']['image']['image']

    images_json = json.loads(list(etcd_node.sudo("crictl images -v -o json").values())[0].stdout)['images']
    for image in images_json:
        if image['id'] == etcd_image_sha:
            return f"{image['repoTags'][0]}"

    raise Exception("Unable to find etcd image on node %s" % node_name)


def export_kubernetes_version(cluster: KubernetesCluster) -> None:
    control_plane = cluster.nodes['control-plane'].get_any_member()
    version = control_plane.sudo('kubectl get nodes --no-headers | head -n 1 | awk \'{print $5; exit}\'').get_simple_out()
    cluster.context['backup_descriptor']['kubernetes']['version'] = version.strip()


def _load_namespaces(logger: log.EnhancedLogger, control_plane: NodeGroup) -> List[Dict[str, str]]:
    namespaces_result = control_plane.sudo(
        'kubectl get ns -o jsonpath=\'{range .items[*]}{.metadata.name}{"\\n"}{end}\'')
    logger.verbose(namespaces_result)
    parsed_namespaces = namespaces_result.get_simple_out().strip().split('\n')
    return [{'name': name} for name in parsed_namespaces]


def _load_resources(logger: log.EnhancedLogger, control_plane: NodeGroup, namespaced: bool) -> List[Dict[str, str]]:
    resources_result = control_plane.sudo(f'kubectl api-resources --verbs=list --sort-by=name '
                                          f'--namespaced{"" if namespaced else "=false"}')
    logger.verbose(resources_result)
    resources_table = utils.parse_aligned_table(resources_result.get_simple_out())

    resources = [
        {
            'name': _resolve_full_resource_name(row['NAME'], row['APIVERSION']),
            'kind': row['KIND'],
            'apiVersion': row['APIVERSION'],
        }
        for row in resources_table
    ]

    return [r for r in resources if r['name'] not in ('events.events.k8s.io', 'events')]


def _resolve_full_resource_name(name: str, apiversion: str) -> str:
    if '/' not in apiversion:
        return name

    return name + '.' + apiversion[0:apiversion.rindex('/')]


def _filter_resources_by_proposed(logger: log.EnhancedLogger,
                                  loaded_resources: List[Dict[str, str]],
                                  proposed_resources: Union[List[str], Literal['all']],
                                  kind: str) -> List[Dict[str, str]]:
    resources = []
    for resource in loaded_resources:
        name = resource['name']
        if proposed_resources == 'all' or name in proposed_resources:
            resources.append(resource)
        else:
            logger.verbose(f'{kind.capitalize()} "{name}" excluded')

    if proposed_resources != 'all':
        for proposed_resource in proposed_resources:
            if not any(proposed_resource == resource['name'] for resource in resources):
                raise Exception(f'Proposed {kind} "{proposed_resource}" not found in loaded cluster {kind}s')

    logger.debug([r['name'] for r in resources])
    return resources


class ParserPayload:
    def __init__(self, event: str, resource_path: str, namespace: Optional[str]):
        self.event = event
        self.resource_path = resource_path
        self.namespace = namespace


class ExportKubernetesParser:
    def __init__(self,
                 logger: log.EnhancedLogger,
                 backup_directory: str,
                 namespaced_resources: List[Dict[str, str]],
                 nonnamespaced_resources: List[Dict[str, str]],
                 graceful_finish_barrier: int,
                 ):
        self.logger = logger
        self.backup_directory = backup_directory
        self.namespaced_resources = namespaced_resources
        self.nonnamespaced_resources = nonnamespaced_resources
        self.namespaced_resources_result_map: Dict[str, List[str]] = {}
        self.nonnamespaced_resources_result: List[str] = []
        self.closed = False
        self.elapsed: float = 0
        self.total_files = 0
        self._graceful_finish_barrier = graceful_finish_barrier
        self._task_queue: 'Queue[ParserPayload]' = Queue()
        self._lock = threading.RLock()

    def schedule(self, payload: ParserPayload) -> None:
        with self._lock:
            if not self.closed:
                self._task_queue.put(payload)
            else:
                self._clear(payload)

    def finish(self, graceful: bool) -> None:
        with self._lock:
            self._graceful_finish_barrier -= 1
            if self.closed:
                return
            if not graceful:
                self._close()
            if self._graceful_finish_barrier == 0 or not graceful:
                self._task_queue.put(ParserPayload("end", "", None))

    def run(self) -> None:
        try:
            while True:
                payload = self._task_queue.get()
                if payload.event == 'end':
                    return

                start = time.time()
                self._handle(payload)
                # do not clear the currently processing payload in case of exception
                self._clear(payload)
                self.elapsed += (time.time() - start)
        finally:
            self._close()

    def _close(self) -> None:
        with self._lock:
            if self.closed:
                return
            # Queue might be not empty in case of exception. Need to clear all pending payloads
            while True:
                try:
                    self._clear(self._task_queue.get_nowait())
                except Empty:
                    break

            self.closed = True

    @staticmethod
    def _clear(payload: ParserPayload) -> None:
        if payload.event == 'do':
            os.remove(payload.resource_path)

    def _parse_identity(self, line: str, api_version: str, kind: str) -> Tuple[str, str]:
        if kind == '':
            kind = self._parse_single_line_property(line, 'kind')
        if api_version == '':
            api_version = self._parse_single_line_property(line, 'apiVersion')

        return api_version, kind

    @staticmethod
    def _parse_single_line_property(line: str, key: str) -> str:
        item_prop = line[2:]
        val = ''
        if item_prop.startswith(f'{key}:'):
            val = yaml.safe_load(item_prop)[key]

        return val

    def _handle(self, payload: ParserPayload) -> None:
        namespace = payload.namespace
        if namespace:
            self.logger.debug(f"Loading resources from namespace {namespace!r}...")
        else:
            self.logger.debug(f"Loading non-namespaced resources...")

        resources = self.nonnamespaced_resources if namespace is None else self.namespaced_resources
        items_by_resource: Dict[str, List[str]] = {}

        def append_item(api_version: str, kind: str, item: str) -> None:
            resource_name = next((
                r['name'] for r in resources
                if r['apiVersion'] == api_version
                and (r['kind'] == kind
                     # TODO remove this W/A for future releases having https://github.com/kubernetes/kubectl/issues/1593 resolved
                     or kind in ('ValidatingAdmissionPolicyList', 'ValidatingAdmissionPolicyBindingList')
                     and r['kind'] == kind[:-len('List')])
            ), None)

            if resource_name is None:
                raise Exception(f"Failed to find resource name for apiVersion: {api_version}, kind: {kind}")

            items_by_resource.setdefault(resource_name, []).append(item)

        with gzip.open(payload.resource_path, 'rt', encoding='utf-8') as file:
            header = ''
            remainder = ''
            item = ''
            api_version, kind = '', ''
            stage = 'header'
            for line in file:
                if stage == 'header':
                    if line.startswith('- '):
                        stage = 'items'
                        item = line
                        api_version, kind = self._parse_identity(line, '', '')
                    else:
                        header += line
                elif stage == 'items':
                    if line.startswith('- '):
                        append_item(api_version, kind, item)
                        item = line
                        api_version, kind = self._parse_identity(line, '', '')
                    elif not line[0].isspace():
                        stage = 'remainder'
                        append_item(api_version, kind, item)
                        remainder = line
                    else:
                        item += line
                        api_version, kind = self._parse_identity(line, api_version, kind)
                elif stage == 'remainder':
                    remainder += line

        for resource in resources:
            resource_name = resource['name']
            items = items_by_resource.get(resource_name)
            if items is None:
                continue

            location = os.path.join(self.backup_directory, 'kubernetes_resources')
            if namespace is not None:
                location = os.path.join(location, namespace)

            os.makedirs(location, exist_ok=True)

            resource_file_path = os.path.join(location, '%s.yaml' % resource_name)
            self.logger.verbose(
                f"Dumping list of resource {resource_name!r}"
                f"{'' if namespace is None else (' for namespace ' + repr(namespace))}...")
            with utils.open_utf8(resource_file_path, 'w') as file:
                file.write(header)
                for item in items:
                    file.write(item)
                file.write(remainder)

            if namespace is None:
                self.nonnamespaced_resources_result.append(resource_name)
            else:
                self.namespaced_resources_result_map.setdefault(namespace, []).append(resource_name)

            self.total_files += 1


class DownloaderPayload:
    def __init__(self, namespace: Optional[str], resources: List[str]):
        self.namespace = namespace
        self.resources = resources


class DownloaderTasksQueue(Iterator[DownloaderPayload]):
    def __init__(self,
                 namespaces: List[Dict[str, str]],
                 namespaced_resources: List[Dict[str, str]],
                 nonnamespaced_resources: List[Dict[str, str]]
                 ):
        self.namespaces = namespaces
        self.namespaced_resources = namespaced_resources
        self.nonnamespaced_resources = nonnamespaced_resources
        self._tasks = self._unsafe_tasks()
        self._lock = threading.Lock()

    def __next__(self) -> DownloaderPayload:
        with self._lock:
            return next(self._tasks)

    def _unsafe_tasks(self) -> Iterator[DownloaderPayload]:
        yield DownloaderPayload(None, [r['name'] for r in self.nonnamespaced_resources])

        for namespace in self.namespaces:
            yield DownloaderPayload(namespace['name'], [r['name'] for r in self.namespaced_resources])


class DownloadException(Exception):
    def __init__(self, task: DownloaderPayload, reason: BaseException):
        self.task = task
        self.reason = reason


class ExportKubernetesDownloader:
    def __init__(self,
                 backup_directory: str,
                 control_plane: NodeGroup,
                 cluster: KubernetesCluster,
                 tasks_queue: Iterator[DownloaderPayload],
                 parser: ExportKubernetesParser,
                 ):
        self.backup_directory = backup_directory
        self.control_plane = control_plane
        self.connection_pool = cluster.create_connection_pool(control_plane.get_hosts())
        # Use own connection instance to run commands
        self.executor = RemoteExecutor(self.control_plane, self.connection_pool)
        self.tasks_queue = tasks_queue
        self.parser = parser
        self.elapsed: float = 0

    def run(self) -> None:
        start = time.time()
        task: Optional[DownloaderPayload] = None
        try:
            while True:
                if self.parser.closed:
                    # No need to continue as the parser is already closed due to error.
                    break

                task = next(self.tasks_queue, None)
                if task is None:
                    break

                # Skip task with empty resource list
                if not task.resources:
                    continue

                random = uuid.uuid4().hex
                temp_local_filepath = os.path.join(self.backup_directory, random)
                self._download(task, temp_local_filepath)

                self.parser.schedule(ParserPayload("do", temp_local_filepath, task.namespace))
        except BaseException as e:
            self.parser.finish(graceful=False)
            if task is not None:
                raise DownloadException(task, e) from None
            else:
                raise
        else:
            self.parser.finish(graceful=True)
        finally:
            self.elapsed = time.time() - start
            self.connection_pool.close()

    def interrupt(self) -> None:
        self.executor.interrupt()

    def _download(self, task: DownloaderPayload, temp_local_filepath: str) -> None:
        namespace = task.namespace
        temp_remote_filepath = utils.get_remote_tmp_path(os.path.basename(temp_local_filepath))

        cmd = f'(set -o pipefail && sudo kubectl ' \
              f'{"" if namespace is None else ("-n " + namespace + " ")}' \
              f'get --ignore-not-found ' \
              f'{",".join(r for r in task.resources)} ' \
              f'-o yaml | gzip -c) > {temp_remote_filepath}'

        group = self.executor.group
        group.run(cmd, pty=True)
        group.get(temp_remote_filepath, temp_local_filepath)
        group.sudo(f'rm {temp_remote_filepath}')
        group.flush()


def export_kubernetes(cluster: KubernetesCluster) -> None:
    backup_directory = prepare_backup_tmpdir(cluster.log, cluster.context)
    control_plane = cluster.nodes['control-plane'].get_any_member()
    backup_kubernetes = cluster.procedure_inventory.get('backup_plan', {}).get('kubernetes', {})
    logger = cluster.log

    logger.debug('Loading namespaces:')
    loaded_namespaces = _load_namespaces(logger, control_plane)
    proposed_namespaces = backup_kubernetes.get('namespaced_resources', {}).get('namespaces', 'all')
    namespaces = _filter_resources_by_proposed(logger, loaded_namespaces, proposed_namespaces, 'namespace')

    logger.debug('Loading namespaced resource types:')
    loaded_resources = _load_resources(logger, control_plane, True)
    proposed_resources = backup_kubernetes.get('namespaced_resources', {}).get('resources', [])
    namespaced_resources = _filter_resources_by_proposed(logger, loaded_resources, proposed_resources, 'resource')

    logger.debug('Loading non-namespaced resource types:')
    loaded_resources = _load_resources(logger, control_plane, False)
    proposed_resources = backup_kubernetes.get('nonnamespaced_resources', [])
    nonnamespaced_resources = _filter_resources_by_proposed(logger, loaded_resources, proposed_resources, 'resource')

    logger.debug('Loading resources:')
    start = time.time()

    # Processing of the particular resource includes
    # 1. downloading of the `kubectl` result (mostly IO)
    # 2. parsing and splitting it into files (can potentially consume CPU)
    control_planes = cluster.nodes['control-plane']
    downloaders_per_control_plane = 2

    parser = ExportKubernetesParser(logger, backup_directory, namespaced_resources, nonnamespaced_resources,
                                    control_planes.nodes_amount() * downloaders_per_control_plane)
    downloader_queue = DownloaderTasksQueue(namespaces, namespaced_resources, nonnamespaced_resources)
    downloaders = [
        ExportKubernetesDownloader(backup_directory, control_plane, cluster,
                                   downloader_queue, parser)
        for _ in range(downloaders_per_control_plane)
        for control_plane in control_planes.get_ordered_members_list()
    ]

    logger.debug(f'Using {len(downloaders)} workers to download resources.')

    with ThreadPoolExecutor(max_workers=len(downloaders) + 1) as tpe:
        parser_async: Optional[concurrent.futures.Future] = None
        downloaders_async: List[concurrent.futures.Future] = []

        def _graceful_shutdown_downloaders() -> Optional[BaseException]:
            # Although parser may exit successfully, not all resources might be processed due to errors in downloaders.
            # Choose any error in such case.
            exc: Optional[BaseException] = None
            for downloader_async in downloaders_async:
                try:
                    downloader_async.result()
                except BaseException as e:
                    logger.verbose(e)
                    if isinstance(e, DownloadException):
                        namespace = e.task.namespace
                        logger.error(f"Failed to download resources {','.join(e.task.resources)}"
                                     + ('' if namespace is None else f" for namespace {namespace!r}."))
                        exc = e.reason
                    else:
                        exc = e

            return exc

        interrupted = False
        exception: Optional[BaseException]
        try:
            while True:
                try:
                    if not interrupted:
                        for downloader in downloaders:
                            downloaders_async.append(tpe.submit(downloader.run))

                        parser_async = tpe.submit(parser.run)
                    elif parser_async is None:
                        break

                    # Wait for successful parsing of resources or till exception.
                    while concurrent.futures.wait({parser_async}, timeout=connections.input_sleep).not_done:
                        pass

                except KeyboardInterrupt:
                    interrupted = True
                    for downloader in downloaders:
                        # Interrupting of downloaders eventually makes them close the parser.
                        # KeyboardInterrupt will be raised from any downloader.
                        downloader.interrupt()

                    continue

                break
        finally:
            # Only obtain the result of finished futures
            try:
                if parser_async is not None:
                    parser_async.result()
            except BaseException as e:
                exception = e
                # Wait for exiting of downloaders after their currently running command is finished / interrupted.
                # If few background tasks fail, the parser exception has priority, and the others are logged.
                _graceful_shutdown_downloaders()
            else:
                exception = _graceful_shutdown_downloaders()

        if isinstance(exception, KeyboardInterrupt):
            raise exception
        elif interrupted:
            raise KeyboardInterrupt()
        elif exception is not None:
            raise exception

    downloaded_resources_descriptor = cluster.context['backup_descriptor']['kubernetes']['resources']
    if parser.namespaced_resources_result_map:
        downloaded_resources_descriptor['namespaced'] = parser.namespaced_resources_result_map

    if parser.nonnamespaced_resources_result:
        downloaded_resources_descriptor['nonnamespaced'] = parser.nonnamespaced_resources_result

    logger.verbose(f'Downloading elapsed: {max(downloader.elapsed for downloader in downloaders)}')
    logger.verbose(f'Parsing elapsed: {parser.elapsed}')
    logger.verbose(f'Total elapsed: {time.time() - start}')
    logger.verbose(f'Total files saved: {parser.total_files}')


def make_descriptor(cluster: KubernetesCluster) -> None:
    backup_directory = prepare_backup_tmpdir(cluster.log, cluster.context)

    cluster.context['backup_descriptor']['kubernetes']['thirdparties'] = cluster.inventory['services']['thirdparties']
    cluster.context['backup_descriptor']['meta']['time']['finished'] = datetime.datetime.now()

    with utils.open_external(os.path.join(backup_directory, 'descriptor.yaml'), 'w') as output:
        output.write(yaml.dump(cluster.context['backup_descriptor']))


def pack_data(cluster: KubernetesCluster) -> None:
    cluster_name = cluster.inventory['cluster_name']
    backup_directory = prepare_backup_tmpdir(cluster.log, cluster.context)

    backup_filename = 'backup-%s-%s.tar.gz' % (cluster_name, utils.get_current_timestamp_formatted())

    target = utils.get_external_resource_path(cluster.procedure_inventory.get('backup_location', backup_filename))
    if os.path.isdir(target):
        target = os.path.join(target, backup_filename)

    cluster.log.debug('Packing all data...')
    pack_to_tgz(target, backup_directory)

    cluster.log.verbose('Cleaning up...')
    shutil.rmtree(backup_directory, ignore_errors=True)


def pack_to_tgz(target_archive: str, source_dir: str) -> None:
    with tarfile.open(target_archive, "w:gz") as tar_handle:
        for root, _, files in os.walk(source_dir):
            for file in files:
                pathname = os.path.join(root, file)
                tar_handle.add(pathname, pathname.replace(source_dir, ''))
        tar_handle.close()


tasks = OrderedDict({
    "verify_backup_location": verify_backup_location,
    "export": {
        "inventory": {
            "cluster_yaml": export_cluster_yaml,
            "ansible_inventory": export_ansible_inventory,
        },
        "lists": {
            "rpms": export_packages_list,
            "hostname": export_hostname,
        },
        "nodes": export_nodes,
        "etcd": export_etcd,
        "cluster_version": export_kubernetes_version,
        "kubernetes": export_kubernetes,
    },
    "make_descriptor": make_descriptor,
    "pack": pack_data,
})


class BackupAction(flow.TasksAction):
    def __init__(self) -> None:
        super().__init__('backup', tasks)


def create_context(cli_arguments: List[str] = None) -> dict:
    cli_help = '''
    Script for making backup of Kubernetes resources and nodes contents.

    How to use:

    '''

    parser = flow.new_procedure_parser(cli_help, tasks=tasks, optional_config=True)

    context = flow.create_context(parser, cli_arguments, procedure='backup')
    context['execution_arguments']['disable_dump'] = False
    context['backup_descriptor'] = {
        'meta': {
            'time': {
                'started': datetime.datetime.now()
            }
        },
        'etcd': {},
        'nodes': {},
        'kubernetes': {
            'resources': {}
        }
    }

    return context


def main(cli_arguments: List[str] = None) -> None:
    context = create_context(cli_arguments)
    flow.ActionsFlow([BackupAction()]).run_flow(context)


if __name__ == '__main__':
    main()



# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import enum
from functools import total_ordering
from typing import Dict, Callable

from kubemarine.core import log
from kubemarine.core.cluster import KubernetesCluster


@total_ordering
class SummaryItem(enum.Enum):
    KUBECONFIG = (0, "Kubeconfig")
    DASHBOARD_URL = (1, "Dashboard URL")
    CONTROL_PLANES = (2, "Running Control Planes")
    WORKERS = (3, "Running Workers")
    ACCOUNT_TOKENS = (4, "Account Tokens File")
    EXECUTION_TIME = (5, "Elapsed")

    def __init__(self, order: int, text: str) -> None:
        self.order = order
        self.text = text

    def __lt__(self, other: 'SummaryItem') -> bool:
        return self.order < other.order


def schedule_report(context: dict, property_: SummaryItem, value: str) -> None:
    context.setdefault('summary_report', {})[property_] = value


def schedule_delayed_report(cluster: KubernetesCluster, call: Callable[[KubernetesCluster], None]) -> None:
    cluster.context.setdefault('delayed_summary_report', []).append(call)
    cluster.schedule_cumulative_point(exec_delayed)


def print_summary(context: dict, logger: log.EnhancedLogger) -> None:
    summary_items: Dict[SummaryItem, str] = context.get('summary_report', {})
    max_length = max(len(si.text) for si in summary_items.keys())
    logger.info('')
    for si, value in sorted(summary_items.items()):
        key = si.text + ': ' + (' ' * (max_length - len(si.text)))
        logger.info(key + value)


def exec_delayed(cluster: KubernetesCluster) -> None:
    for call in cluster.context.get('delayed_summary_report', []):
        call(cluster)



# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import sys
from typing import List, OrderedDict, Any, Protocol, runtime_checkable

NO_TASKS_PROCEDURES = ["do", "migrate_kubemarine", "config"]


@runtime_checkable
class Procedure(Protocol):
    def create_context(self, cli_arguments: List[str] = None) -> dict: ...

    def main(self, cli_arguments: List[str] = None) -> Any: ...


@runtime_checkable
class TasksProcedure(Procedure, Protocol):
    @property
    def tasks(self) -> OrderedDict[str, Any]: ...


def import_procedure(name: str) -> Procedure:
    if name in NO_TASKS_PROCEDURES:
        return _import_no_tasks_procedure(name)

    return _import_tasks_procedure(name)


def _import_tasks_procedure(name: str) -> TasksProcedure:
    procedure: TasksProcedure
    if name == "add_node":
        from kubemarine.procedures import add_node as procedure
    elif name == "backup":
        from kubemarine.procedures import backup as procedure
    elif name == "cert_renew":
        from kubemarine.procedures import cert_renew as procedure
    elif name == "check_iaas":
        from kubemarine.procedures import check_iaas as procedure
    elif name == "check_paas":
        from kubemarine.procedures import check_paas as procedure
    elif name == "install":
        from kubemarine.procedures import install as procedure
    elif name == "manage_pss":
        from kubemarine.procedures import manage_pss as procedure
    elif name == "reboot":
        from kubemarine.procedures import reboot as procedure
    elif name == "reconfigure":
        from kubemarine.procedures import reconfigure as procedure
    elif name == "remove_node":
        from kubemarine.procedures import remove_node as procedure
    elif name == "restore":
        from kubemarine.procedures import restore as procedure
    elif name == "upgrade":
        from kubemarine.procedures import upgrade as procedure
    else:
        raise NotImplementedError(f"Procedure {name!r} is not implemented yet")

    return procedure


def _import_no_tasks_procedure(name: str) -> Procedure:
    procedure: Procedure
    if name == "do":
        from kubemarine.procedures import do as procedure
    elif name == "migrate_kubemarine":
        from kubemarine.procedures import migrate_kubemarine as procedure
    elif name == "config":
        from kubemarine.procedures import config as procedure
    else:
        raise NotImplementedError(f"Procedure {name!r} is not implemented yet")

    return procedure


def deimport_procedure(name: str) -> None:
    del sys.modules['kubemarine.procedures.%s' % name]



# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Callable, TypeVar

from typing_extensions import ParamSpec

from kubemarine.core.group import NodeGroup


_P = ParamSpec('_P')
_T = TypeVar('_T')


def restrict_multi_os_group(fn: Callable[_P, _T]) -> Callable[_P, _T]:
    """
    Method is an annotation that does not allow origin method to use different OS families in the same group.
    :param fn: Origin function to apply annotation validation to
    :return: Validation wrapper function
    """
    def wrapper(*args: _P.args, **kwargs: _P.kwargs) -> _T:
        group = next((g for g in args if isinstance(g, NodeGroup)), None)
        if group is None:
            raise Exception("Failed to find argument of NodeGroup type")
        if group.is_multi_os():
            raise Exception(f'Method "{str(fn)}" do not supports multi-os group')
        return fn(*args, **kwargs)
    return wrapper


def restrict_empty_group(fn: Callable[_P, _T]) -> Callable[_P, _T]:
    """
    Method is an annotation that prohibits passing empty groups to the function.
    :param fn: Origin function to apply annotation validation to
    :return: Validation wrapper function
    """
    def wrapper(*args: _P.args, **kwargs: _P.kwargs) -> _T:
        group = next((g for g in args if isinstance(g, NodeGroup)), None)
        if group is None:
            raise Exception("Failed to find argument of NodeGroup type")
        if group.is_empty():
            raise Exception(f'Method "{str(fn)}" prohibits passing empty groups to it')
        return fn(*args, **kwargs)
    return wrapper



# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Tuple, Optional

from deepmerge import Merger  # type: ignore[import-untyped]


def is_list_extends(nxt: list, path: list) -> bool:
    return get_strategy_position(nxt, path)[0] == 'merge'


def get_strategy_position(nxt: list, path: list) -> Tuple[Optional[str], int]:
    strategy = None
    strategy_definition_position = 0
    for i, v in enumerate(nxt):
        if isinstance(v, dict) and '<<' in v:
            if strategy is not None:
                raise Exception(f"Found more than one merge strategy definitions at path {path}.")
            strategy = v.get('<<')
            strategy_definition_position = i
            if v.keys() != {'<<'} or strategy not in ('replace', 'merge'):
                raise Exception(f"Unexpected merge strategy definition {v} at path {path}.")

    return strategy, strategy_definition_position


def list_merger(_: Merger, path: list, base: list, nxt: list) -> list:
    strategy, strategy_definition_position = get_strategy_position(nxt, path)

    if strategy is None:
        return nxt

    elements_after = nxt[(strategy_definition_position + 1):]
    elements_before = nxt[:strategy_definition_position]
    # do not modify source list
    nxt = []
    nxt.extend(elements_before)
    if strategy == 'merge':
        nxt.extend(base)
    nxt.extend(elements_after)

    return nxt


default_merger: Merger = Merger(
    [
        (list, [list_merger]),
        (dict, ["merge"])
    ],
    ["override"],
    ["override"]
)

override_merger: Merger = Merger(
    [
        (list, ["override"]),
        (dict, ["merge"])
    ],
    ["override"],
    ["override"]
)



# Copyright 2021-2023 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
from abc import ABC, abstractmethod
from typing import Any, Union, Sequence, List, Iterator, Mapping, Type, Dict

import yaml

Index = Union[str, int]
Primitive = Union[str, int, bool, float]


class Proxy(ABC):
    """
    Abstract class that proxies getter methods to some other container.
    """

    def __repr__(self) -> str:
        return repr(self._KM_materialize())

    @abstractmethod
    def _KM_materialize(self) -> Any:
        """
        Create real container that can be serialized.
        This method triggers access to all the items in the proxied container.

        :return: real container that can be serialized.
        """
        pass


class DelegatingProxy(Proxy, ABC):
    """
    Abstract class that proxies getter methods to some real container.
    """

    @abstractmethod
    def _KM_unsafe(self) -> Union[dict, list]:
        """
        Real proxied `dict` or `list`. It can be not the same as the proxied container.
        Should be used only to obtain sequence of indexes (`str` keys of `dict`, or `int` indexes of `list`).

        :return: read proxied `dict` or `list`.
        """
        pass

    @abstractmethod
    def _KM__getitem__(self, index: Index) -> Union[Primitive, Proxy]:
        """
        Get item from `dict` by string key, or from `list` by integer index.
        If the item is also a container, it should be proxied.

        :param index: key of either `str`, or `int` type.
        :return: primitive value or Proxy
        """
        pass


class MappingProxy(DelegatingProxy, Mapping[str, Any], ABC):
    """
    The main facade that proxies all mapping methods to some real mapping.
    """

    def __getitem__(self, k: str) -> Any:
        return self._KM__getitem__(k)

    def __len__(self) -> int:
        return len(self._KM_unsafe())

    def __iter__(self) -> Iterator[str]:
        return iter(self._KM_unsafe())

    def _KM_materialize(self) -> Any:
        return dict(self)


class SequenceProxy(DelegatingProxy, Sequence[Any], ABC):
    """
    The main facade that proxies all sequence methods to some real sequence.
    """

    def __getitem__(self, index: Union[int, slice]) -> Any:
        if isinstance(index, slice):
            indexes = list(range(len(self)))[index]
            return SliceProxy(self, indexes)

        return self._KM__getitem__(index)

    def __len__(self) -> int:
        return len(self._KM_unsafe())

    def _KM_materialize(self) -> Any:
        return list(self)


class SliceProxy(Proxy, Sequence[Any]):
    """
    Implementation of `slice` over SequenceProxy.
    """

    def __init__(self, sequence: SequenceProxy, indexes: List[int]):
        self._KM_sequence = sequence
        self._KM_indexes = indexes

    def __getitem__(self, index: Union[int, slice]) -> Any:
        if isinstance(index, slice):
            indexes = self._KM_indexes[index]
            return SliceProxy(self._KM_sequence, indexes)

        return self._KM_sequence[self._KM_indexes[index]]

    def __len__(self) -> int:
        return len(self._KM_indexes)

    def _KM_materialize(self) -> Any:
        return list(self)


class Node:
    """
    Highly extendable wrapper over `dict` or `list`.
    """

    def __init__(self, delegate: Union[dict, list]):
        self.delegate = delegate

    def descend(self, index: Index) -> Union[Primitive, 'Node']:
        """
        The main custom implementation to access the items of real `dict` or `list` from Proxies.

        Get a child item / value by the specified `index` from the underlying container.

        If the child item is also a `dict` or `list`, wrap it with the new instance of `Node`
        with (probably) its own logic to resolve items.

        If the child item is a primitive value, it can be arbitrarily converted.

        :param index: key of either `str`, or `int` type.
        :return: primitive value or new child Node instance.
        """
        val: Union[Primitive, list, dict] = self.delegate[index]  # type: ignore[index]
        if isinstance(val, (list, dict)):
            return self._child(index, val)

        return val

    def _child(self, index: Index, val: Union[list, dict]) -> 'Node':
        """
        Instantiate new Node wrapper over the specified child item `val`.
        It can be overridden if derived class has additional constructor parameters,
        but should always instantiate an instance of `_child_type`.

        :param index: key of either `str`, or `int` type.
        :param val: wrapped `dict` or `list`.
        :return: new child Node.
        """
        return self._child_type(index)(val)

    def _child_type(self, _: Index) -> Type['Node']:
        """
        Return `Node` class that should wrap a child item by the specified `index`.

        To preserve behaviour of parent `Node`, the method should be overridden to return the derived `Node` class.

        :return: `Node` class to wrap child item.
        """
        return Node


class MutableNode(Node, ABC):
    """
    A Node that can change underlying `dict` or `list` on-the-fly during access to its items.
    """

    def descend(self, index: Index) -> Union[Primitive, Node]:
        child = super().descend(index)
        if isinstance(child, Node):
            return child

        val = self._convert(index, child)
        self.delegate[index] = val  # type: ignore[index]

        return val

    @abstractmethod
    def _convert(self, index: Index, val: Primitive) -> Primitive:
        """
        Convert primitive value before putting to the underlying container.

        :param index: key of either `str`, or `int` type.
        :param val: primitive value
        :return: converted value
        """
        pass

    def _child_type(self, _: Index) -> Type[Node]:
        return MutableNode


class NodeProxy(DelegatingProxy, ABC):
    """
    Abstract class that proxies getter methods to an instance of `Node`.
    """

    def __init__(self, node: Node):
        self._KM_node = node
        self._KM_cached: Dict[Index, Union[Primitive, Proxy]] = {}

    def _KM_unsafe(self) -> Union[dict, list]:
        return self._KM_node.delegate

    def _KM__getitem__(self, index: Index) -> Union[Primitive, Proxy]:
        if index not in self._KM_cached:
            child = self._KM_node.descend(index)
            val = ((NodeMapping(child) if isinstance(child.delegate, dict) else NodeSequence(child))
                   if isinstance(child, Node)
                   else child)

            self._KM_cached[index] = val
            return val

        return self._KM_cached[index]


class NodeMapping(NodeProxy, MappingProxy):
    """
    Mapping that proxies all getter methods to an instance of `Node`.
    """
    pass


class NodeSequence(NodeProxy, SequenceProxy):
    """
    Sequence that proxies all getter methods to an instance of `Node`.
    """
    pass


class ProxyJSONEncoder(json.JSONEncoder):
    """
    Supports serialization of any Proxy to JSON.
    """

    def default(self, o: Any) -> Any:
        if isinstance(o, Proxy):
            return o._KM_materialize()  # pylint: disable=protected-access

        return super().default(o)


def proxy_representer(dumper: yaml.Dumper, data: Proxy) -> Any:
    val = data._KM_materialize()  # pylint: disable=protected-access
    return dumper.yaml_representers[type(val)](dumper, val)


class ProxyDumper(yaml.Dumper):  # pylint: disable=too-many-ancestors
    """
    Supports serialization of any `Node` proxy to YAML.
    """
    pass


ProxyDumper.add_representer(NodeMapping, proxy_representer)
ProxyDumper.add_representer(NodeSequence, proxy_representer)
ProxyDumper.add_representer(SliceProxy, proxy_representer)



# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import io
import re
from typing import Tuple, Dict, List

from kubemarine.core import utils, log
from kubemarine.core.group import NodeGroup, RunnersGroupResult

# Common regexp should support the following schemes:
# SELinux status:                 enabled
# SELinuxfs mount:                /selinux
# Policy version:                 23

# Commong regexp
common_regexp = "%s:\\s*([\\w/\\d]*)"

# Structure with names to parse and keys to put the data to parsed map:
parsed_names_map = {
    re.compile(common_regexp % 'SELinux status', re.M):             'status',
    re.compile(common_regexp % 'SELinuxfs mount', re.M):            'mount',
    re.compile(common_regexp % 'SELinux root directory', re.M):     'root_directory',
    re.compile(common_regexp % 'Current mode', re.M):               'mode',
    re.compile(common_regexp % 'Mode from config file', re.M):      'mode_from_file',
    re.compile(common_regexp % 'Loaded policy name', re.M):         'policy',
    re.compile(common_regexp % 'Policy from config file', re.M):    'policy_from_file',
    re.compile(common_regexp % 'Policy MLS status', re.M):          'policy_mls',
    re.compile(common_regexp % 'Policy deny_unknown status', re.M): 'policy_deny_unknown',
    re.compile(common_regexp % 'Policy version', re.M):             'policy_version',
    re.compile(common_regexp % 'Max kernel policy version', re.M):  'policy_version_max'
}

# Cutomized permissive types regexp
# It should from the following:
#
# > Customized Permissive Types
# >
# > keepalived_t
# > something_else_t
# >
# > Builtin Permissive Types
# >
# > and_another_one
#
# Cut only the following:
#
# > keepalived_t
# > something_else_t
#
permissive_types_regex = re.compile("Customized Permissive Types\\s*([\\w_\\s]*)\\s*", re.M)


def get_expected_state(inventory: dict) -> str:
    state: str = inventory['services']['kernel_security'].get('selinux', {}).get('state', 'enforcing')
    return state


def get_expected_policy(inventory: dict) -> str:
    policy: str = inventory['services']['kernel_security'].get('selinux', {}).get('policy', 'targeted')
    return policy


def get_expected_permissive(inventory: dict) -> List[str]:
    permissive: List[str] = inventory['services']['kernel_security'].get('selinux', {}).get('permissive', [])
    return permissive


def parse_selinux_status(logger: log.EnhancedLogger, stdout: str) -> Dict[str, str]:
    result: Dict[str, str] = {}
    if stdout is not None and stdout.strip() != '':
        for regex, key in parsed_names_map.items():
            matches: List[str] = re.findall(regex, stdout)
            if matches:
                result[key] = matches[0].strip()
    logger.verbose('Parsed status: %s' % result)
    return result


def parse_selinux_permissive_types(logger: log.EnhancedLogger, stdout: str) -> List[str]:
    if stdout is None or stdout.strip() == '':
        logger.verbose('Permissive types pattern not found - presented stdout is empty')
        return []

    matches: List[str] = re.findall(permissive_types_regex, stdout)
    if not matches:
        logger.verbose('Permissive types pattern not found')
        return []

    types_string = matches[0]
    if types_string.strip() == '':
        logger.verbose('Permissive types pattern found, but value is empty')
        return []

    result = types_string.split('\n')
    logger.verbose('Permissive types parsed: %s' % result)
    return result


def get_selinux_status(group: NodeGroup) -> Tuple[RunnersGroupResult, Dict[str, dict]]:
    log = group.cluster.log

    result = group.sudo("sestatus && sudo semanage permissive -l")

    parsed_result: Dict[str, dict] = {}
    for host, node_result in result.items():
        log.verbose('Parsing status for %s...' % host)
        parsed_result[host] = parse_selinux_status(log, node_result.stdout)
        parsed_result[host]['permissive_types'] = parse_selinux_permissive_types(log, node_result.stdout)
    log.verbose("Parsed remote sestatus summary:\n%s" % parsed_result)
    return result, parsed_result


def is_config_valid(group: NodeGroup, state: str = None, policy: str = None, permissive: List[str] = None) \
        -> Tuple[bool, RunnersGroupResult, Dict[str, dict]]:
    log = group.cluster.log

    if group.get_nodes_os() == 'debian':
        raise Exception("Selinux is not supported on Ubuntu/Debian os family")

    log.verbose('Verifying selinux configs...')

    if state is None:
        state = get_expected_state(group.cluster.inventory)

    if policy is None:
        policy = get_expected_policy(group.cluster.inventory)

    if permissive is None:
        permissive = get_expected_permissive(group.cluster.inventory)

    result, parsed_result = get_selinux_status(group)
    valid = True

    for host, selinux_status in parsed_result.items():

        if selinux_status['status'] == 'disabled' and state == 'disabled':
            continue

        # for some different selinux versions some statuses may be absent
        # that is why such construction was made - when no status, then cause true
        if state != selinux_status.get('mode', state) or \
                state != selinux_status.get('mode_from_file', state) or \
                policy != selinux_status.get('policy_from_file', policy) or \
                policy != selinux_status.get('policy', policy):
            valid = False
            log.verbose('Selinux configs are not matched at %s' % host)
            break

        if permissive:
            for permissive_type in permissive:
                if permissive_type not in selinux_status['permissive_types']:
                    valid = False
                    log.verbose('Permissive type %s not found in types %s at %s '
                                % (permissive_type, selinux_status['permissive_types'], host))
                    break
            # if no break was called in previous for loop, then else called and no break will be called in current loop
            else:
                continue
            # if break was called in previous for loop, then do break in current loop
            break

    return valid, result, parsed_result


def setup_selinux(group: NodeGroup) -> bool:
    log = group.cluster.log

    # this method handles cluster with multiple os, suppressing should be enabled
    if group.get_nodes_os() not in ['rhel', 'rhel8', 'rhel9']:
        log.debug("Skipped - selinux is not supported on Ubuntu/Debian os family")
        return False

    expected_state = get_expected_state(group.cluster.inventory)
    expected_policy = get_expected_policy(group.cluster.inventory)
    expected_permissive = get_expected_permissive(group.cluster.inventory)

    valid, result, _ = \
        is_config_valid(group,
                        state=expected_state,
                        policy=expected_policy,
                        permissive=expected_permissive)

    if valid:
        log.debug("Skipped - selinux already correctly configured")
        log.debug(result)
        return False

    config = io.StringIO('SELINUX=%s\nSELINUXTYPE=%s\n' % (expected_state, expected_policy))

    log.debug("Uploading selinux config...")
    utils.dump_file(group.cluster, config, 'selinux_config')
    group.put(config, '/etc/selinux/config', backup=True, sudo=True)

    semanage_commands = ''
    for item in expected_permissive:
        if semanage_commands != '':
            semanage_commands = semanage_commands + ' && sudo '
        semanage_commands = semanage_commands + 'semanage permissive -a %s' % item
    log.verbose("The following command will be executed to configure permissive:\n%s" % semanage_commands)

    group.sudo(semanage_commands)

    return True



# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
All files and directories inside kubemarine/patches directory should participate only in patching mechanism,
and relate to the current Kubemarine version.

The whole directory is automatically cleared and reset after new version of Kubemarine is released.
"""

from typing import List

from kubemarine.core.patch import Patch

patches: List[Patch] = [
]
"""
List of patches that is sorted according to the Patch.priority() before execution.
Patches that have the same priority, are executed in the declared order.
"""



# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import io
from typing import Tuple, Optional, Dict, List, Union

from kubemarine import jinja
from kubemarine.core import utils, static, errors
from kubemarine.core.cluster import KubernetesCluster, EnrichmentStage, enrichment
from kubemarine.core.group import NodeGroup, RunnersGroupResult
from kubemarine.core.yaml_merger import default_merger

ERROR_SHA1_NOT_CHANGED = (
    "Key 'source' was changed for third-party {thirdparty!r} during upgrade {previous_version} -> {version}, "
    "but 'sha1' was not changed.")


def is_default_thirdparty(destination: str) -> bool:
    return destination in static.GLOBALS['thirdparties']


def get_default_thirdparties() -> List[str]:
    return list(static.GLOBALS['thirdparties'])


def get_default_thirdparty_version(kubernetes_version: str, destination: str) -> str:
    """
    :param kubernetes_version: Kubernetes version
    :param destination: absolute path of default third-party
    :return: version of third-party from compatibility map
    """
    software_settings = _get_software_settings_for_thirdparty(kubernetes_version, destination)

    if 'version' in software_settings:
        return software_settings['version']
    else:
        # kubeadm, kubelet, kubectl
        return kubernetes_version


def get_default_thirdparty_source(destination: str, version: str, in_public: bool) -> str:
    """
    :param destination: absolute path of default third-party
    :param version: version of third-party
    :param in_public: flag whether to return third-party URL in public resources.
    :return: URL of the third-party source.
    """
    if not is_default_thirdparty(destination):
        raise Exception(f"{destination} is not a default 3rd-party")

    thirdparty_settings = static.GLOBALS['thirdparties'][destination]
    if in_public:
        source_prefix = thirdparty_settings['source_prefix']['public']
    else:
        source_prefix = thirdparty_settings['source_prefix']['private']

    relative_path = thirdparty_settings['relative_path'].format(version=version)
    return f"{source_prefix}/{relative_path}"


def get_default_thirdparty_identity(inventory: dict,
                                    destination: str, in_public: bool) -> Tuple[str, str]:
    """
    :param inventory: inventory of the cluster
    :param destination: absolute path of default third-party
    :param in_public: flag whether to return third-party URL in public resources.
    :return: a pair of the third-party URL and sha1
    """
    kubernetes_version = inventory['services']['kubeadm']['kubernetesVersion']

    software_settings = _get_software_settings_for_thirdparty(kubernetes_version, destination)
    sha1 = software_settings['sha1']

    version = get_default_thirdparty_version(kubernetes_version, destination)
    source = get_default_thirdparty_source(destination, version, in_public)

    return source, sha1


def _get_software_settings_for_thirdparty(kubernetes_version: str, destination: str) -> Dict[str, str]:
    if not is_default_thirdparty(destination):
        raise Exception(f"{destination} is not a default 3rd-party")

    thirdparty_settings = static.GLOBALS['thirdparties'][destination]
    software_name = thirdparty_settings['software_name']
    software_settings: Dict[str, str] = static.GLOBALS['compatibility_map']['software'][software_name][kubernetes_version]
    return software_settings


def get_thirdparty_destination(software_name: str) -> str:
    destination: str
    for destination, thirdparty_settings in static.GLOBALS['thirdparties'].items():
        if thirdparty_settings['software_name'] == software_name:
            return destination

    raise Exception(f"Failed to find third-party destination for {software_name!r}")


def get_thirdparty_recommended_sha(destination: str, cluster: KubernetesCluster) -> Optional[str]:
    if not is_default_thirdparty(destination):
        # 3rd-party is not managed by Kubemarine
        return None

    cluster.log.verbose("Calculate recommended sha for thirdparty %s..." % destination)
    _, recommended_sha = get_default_thirdparty_identity(cluster.inventory,
                                                         destination, in_public=True)
    cluster.log.verbose(f"Recommended sha for thirdparty {destination} was calculated: {recommended_sha}")

    return recommended_sha


def _convert_thirdparty(thirdparties: dict, destination: str) -> dict:
    config: Union[str, dict] = thirdparties.setdefault(destination, {})
    if isinstance(config, str):
        thirdparties[destination] = config = {
            'source': config
        }

    return config


def _get_procedure_plan(cluster: KubernetesCluster) -> List[Tuple[str, dict]]:
    context = cluster.context
    procedure = context["initial_procedure"]
    procedure_inventory = cluster.procedure_inventory
    upgrade_plan = []
    if procedure == "upgrade":
        kubernetes_version = cluster.inventory['services']['kubeadm']['kubernetesVersion']
        for i, v in enumerate(procedure_inventory['upgrade_plan'][context['upgrade_step']:]):
            version = kubernetes_version if i == 0 else v
            upgrade_plan.append((version, procedure_inventory.get(v, {}).get("thirdparties", {})))

    elif procedure == "migrate_kubemarine" and 'upgrading_thirdparty' in context:
        procedure_thirdparties = procedure_inventory.get('upgrade', {}).get("thirdparties", {})
        procedure_thirdparties = utils.subdict_yaml(procedure_thirdparties, [context['upgrading_thirdparty']])
        upgrade_plan = [("", procedure_thirdparties)]
    elif procedure == "restore":
        upgrade_plan = [("", procedure_inventory.get('restore_plan', {}).get('thirdparties', {}))]

    return upgrade_plan


@enrichment(EnrichmentStage.PROCEDURE, procedures=['upgrade', 'migrate_kubemarine', 'restore'])
def enrich_procedure_inventory(cluster: KubernetesCluster) -> None:
    procedure_plan = _get_procedure_plan(cluster)
    procedure_thirdparties = {} if not procedure_plan else procedure_plan[0][1]

    if procedure_thirdparties:
        thirdparties = cluster.inventory.setdefault("services", {}).setdefault("thirdparties", {})
        procedure_thirdparties = utils.deepcopy_yaml(procedure_thirdparties)

        for destination in procedure_thirdparties:
            config = _convert_thirdparty(thirdparties, destination)
            procedure_config = _convert_thirdparty(procedure_thirdparties, destination)
            default_merger.merge(config, procedure_config)


@enrichment(EnrichmentStage.PROCEDURE, procedures=['upgrade', 'migrate_kubemarine'])
def verify_procedure_inventory(cluster: KubernetesCluster) -> None:
    context = cluster.context

    upgrade_plan = _get_procedure_plan(cluster)
    if not upgrade_plan:
        return

    if context["initial_procedure"] == "upgrade":
        previous_version = cluster.previous_inventory['services']['kubeadm']['kubernetesVersion']
        thirdparties_verify = get_default_thirdparties()
    else:  # migrate_kubemarine procedure
        previous_version = ""
        thirdparties_verify = [context['upgrading_thirdparty']]

    _verify_upgrade_plan(cluster, previous_version, thirdparties_verify, upgrade_plan)


def _verify_upgrade_plan(cluster: KubernetesCluster, previous_version: str,
                         thirdparties_verify: List[str], upgrade_plan: List[Tuple[str, dict]]) -> None:

    previous_raw_thirdparties = utils.deepcopy_yaml(cluster.previous_raw_inventory.get('services', {}).get('thirdparties', {}))
    sensitive_keys = ['source', 'sha1']

    # Restrict validation to the next version only because it depends on compilation.
    for version, upgrade_thirdparties in upgrade_plan[:1]:
        if version != '' and jinja.is_template(version):
            break

        upgrade_thirdparties = utils.deepcopy_yaml(upgrade_thirdparties)

        for destination in thirdparties_verify:
            raw_config = _convert_thirdparty(previous_raw_thirdparties, destination)
            # No need to copy two below inventories as they are (to be) enriched.
            previous_config = _convert_thirdparty(cluster.previous_inventory['services']['thirdparties'], destination)
            config = _convert_thirdparty(cluster.inventory['services']['thirdparties'], destination)

            upgrade_config = _convert_thirdparty(upgrade_thirdparties, destination)

            for key in sensitive_keys:
                defaults_changed = True if version == '' else (
                        get_default_thirdparty_version(previous_version, destination)
                        != get_default_thirdparty_version(version, destination)
                )

                if raw_config.get(key) and (
                        not upgrade_config.get(key)
                        # It is historically possible to redefine all thirdparties with templates
                        # dependent on Kubernetes version and the compatibility map.
                        # This technically allows to do not supply new thirdparties during upgrade, but still upgrade them.
                        # The validation below checks this case and allows upgrade if defaults are changed,
                        # and the result of compilation differs.
                        # For Kubemarine migration procedure, the check below is always True.
                        and defaults_changed and previous_config.get(key) == config.get(key)
                ):
                    raise errors.KME("KME0011",
                                     key=key, thirdparty=destination,
                                     previous_version_spec=f" for version {previous_version}" if previous_version else "",
                                     next_version_spec=f" for next version {version}" if version else "")

            if (config.get('source') and previous_config['source'] != config['source']
                    and 'sha1' in previous_config and 'sha1' in config
                    and previous_config['sha1'] == config['sha1']):
                raise Exception(ERROR_SHA1_NOT_CHANGED.format(
                    thirdparty=destination, previous_version=previous_version, version=version))

            default_merger.merge(raw_config, upgrade_config)

        previous_version = version


@enrichment(EnrichmentStage.FULL)
def enrich_inventory_apply_defaults(cluster: KubernetesCluster) -> None:
    inventory = cluster.inventory
    thirdparties: Dict[str, dict] = inventory['services']['thirdparties']

    for destination in thirdparties:

        config = _convert_thirdparty(thirdparties, destination)

        if config.get('mode') is None:
            config['mode'] = 700

        if config.get('owner') is None:
            config['owner'] = 'root'

        if config.get('group') is not None:
            config['groups'] = [config['group']]
            del config['group']

        if config.get('node') is not None:
            config['nodes'] = [config['node']]
            del config['node']

        if config.get('groups') is None and config.get('nodes') is None:
            config['groups'] = ['control-plane', 'worker']

        if config.get('nodes') is not None:
            all_nodes_names = cluster.nodes['all'].get_nodes_names()
            for node_name in config['nodes']:
                if node_name not in all_nodes_names:
                    raise Exception('Unknown node name provided for thirdparty %s. '
                                    'Expected any of %s, but \'%s\' found.'
                                    % (destination, all_nodes_names, node_name))

        if is_default_thirdparty(destination) and 'source' not in config:
            source, sha1 = get_default_thirdparty_identity(inventory, destination, in_public=True)
            config['source'] = source
            if 'sha1' not in config:
                config['sha1'] = sha1


def get_install_group(cluster: KubernetesCluster, config: dict) -> NodeGroup:
    return cluster.create_group_from_groups_nodes_names(
        config.get('groups', []), config.get('nodes', []))


def get_group_require_unzip(cluster: KubernetesCluster, inventory: dict) -> NodeGroup:
    thirdparties: dict = inventory['services']['thirdparties']

    group = cluster.make_group([])
    for destination, config in thirdparties.items():
        extension = destination.split('.')[-1]
        if config.get('unpack') is None or extension != 'zip':
            continue

        install_group = get_install_group(cluster, config)
        group = group.include_group(install_group)

    return group


def install_thirdparty(filter_group: NodeGroup, destination: str) -> Optional[RunnersGroupResult]:
    cluster = filter_group.cluster
    config = cluster.inventory['services'].get('thirdparties', {}).get(destination)

    if config is None:
        raise Exception('Not possible to install thirdparty %s - not found in configfile' % destination)

    common_group = get_install_group(cluster, config)
    common_group = common_group.intersection_group(filter_group)

    if common_group.is_empty():
        cluster.log.verbose(f'No destination nodes to install thirdparty {destination!r}')
        return None

    cluster.log.debug("Thirdparty \"%s\" will be installed" % destination)
    is_curl = config['source'][:4] == 'http' and '://' in config['source'][4:8]

    # all commands will be grouped to single run
    remote_commands = ''

    # directory will be created if it is not exists
    destination_directory = '/'.join(destination.split('/')[:-1])
    cluster.log.verbose('Destination directory: %s' % destination_directory)

    # ! In the further code there is no error and nothing is missing !
    # Here a long shell command is intentionally constructed and executed at once to speed up work
    # At the same time, in the middle of the construction of the command, a file may suddenly be uploaded and then
    # the command will be executed in two runs instead of single run

    # is destination directory exists?
    remote_commands += 'mkdir -p %s' % destination_directory

    if config.get('sha1') is not None:
        cluster.log.verbose('SHA1 hash is defined, it will be used during installation')

    if is_curl:
        cluster.log.verbose('Installation via curl download detected')
        if config.get('sha1') is not None:
            # if hash equal, then stop further actions immediately! unpack should not be performed too
            remote_commands += (' && [ -f %s ] && [ "%s" == "$(sudo openssl sha1 %s | sed "s/^.* //")" ]'
                                ' && exit 0 || true '
                                % (destination, config['sha1'], destination))
        remote_commands += (' && sudo rm -f %s && sudo curl --max-time %d -k -f -g -s --show-error -L %s -o %s && '
                            % (destination, cluster.inventory['globals']['timeout_download'], config['source'], destination))
    else:
        cluster.log.verbose('Installation via sftp upload detected')
        if config.get('sha1') is not None:
            remote_commands += (' && [ -f %s ] && sudo openssl sha1 %s | sed "s/^.* //" || true'
                                % (destination, destination))

        result = common_group.sudo(remote_commands)
        group_to_upload = common_group
        compare_hashes = True
        if config.get('sha1') is not None:
            group_to_upload = common_group.exclude_group(result.get_nodes_group_where_value_in_stdout(config['sha1']))
            compare_hashes = False

        remote_commands = ''
        script = utils.read_internal(config['source'])
        group_to_upload.put(io.StringIO(script), destination, sudo=True, compare_hashes=compare_hashes)

    remote_commands += 'sudo chmod %s %s' % (config['mode'], destination)
    remote_commands += ' && sudo chown %s %s' % (config['owner'], destination)
    remote_commands += ' && sudo ls -la %s' % destination

    if config.get('unpack') is not None:
        cluster.log.verbose('Unpack request detected')

        remote_commands += ' && sudo mkdir -p %s' % config['unpack']

        extension = destination.split('.')[-1]
        if extension == 'zip':
            cluster.log.verbose('Unzip will be used for unpacking')
            remote_commands += ' && sudo unzip -o %s -d %s' % (destination, config['unpack'])
            
            remote_commands += ' && sudo unzip -qq -l %s | awk \'NF > 3 { print $4 }\'| xargs -I FILE sudo chmod %s %s/FILE' \
                   % (destination, config['mode'], config['unpack'])
            remote_commands += ' && sudo unzip -qq -l %s | awk \'NF > 3 { print $4 }\'| xargs -I FILE sudo chown -R %s %s/FILE' \
                   % (destination, config['owner'], config['unpack'])
            remote_commands += ' && sudo unzip -qq -l %s | awk \'NF > 3 { print $4 }\'| xargs -I FILE sudo ls -la %s/FILE' \
                   % (destination, config['unpack'])
            
        else:
            cluster.log.verbose('Tar will be used for unpacking')
            remote_commands += ' && sudo tar -zxf %s -C %s' % (destination, config['unpack'])
            
            remote_commands += ' && sudo tar -tf %s | xargs -I FILE sudo chmod %s %s/FILE' \
                           % (destination, config['mode'], config['unpack'])
            remote_commands += ' && sudo tar -tf %s | xargs -I FILE sudo chown %s %s/FILE' \
                           % (destination, config['owner'], config['unpack'])
            remote_commands += ' && sudo tar -tf %s | xargs -I FILE sudo ls -la %s/FILE' % (destination, config['unpack'])

    return common_group.sudo(remote_commands, pty=True)


def install_all_thirparties(group: NodeGroup) -> None:
    cluster: KubernetesCluster = group.cluster
    log = cluster.log

    if not cluster.inventory['services'].get('thirdparties', {}):
        return

    for destination in cluster.inventory['services']['thirdparties'].keys():
        managing_plugin: Optional[str] = None

        # install and upgrade procedures have separate tasks for thirdparties managed by plugins
        if cluster.context.get("initial_procedure") in ("install", "upgrade"):
            managing_plugin = next((plugin_name
                                    for plugin_name, plugin_configs in cluster.inventory['plugins'].items()
                                    for plugin_procedure in plugin_configs.get('installation', {}).get('procedures', [])
                                    if plugin_procedure.get('thirdparty') == destination),
                                   None)

        if managing_plugin is not None:
            log.verbose('Thirdparty \'%s\' installation is delayed as it should be installed with \'%s\' plugin.'
                        % (destination, managing_plugin))
        else:
            res = install_thirdparty(group, destination)
            if res is not None:
                log.debug(res)



#!/usr/bin/env python3
# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from collections import OrderedDict
from typing import List

from kubemarine import kubernetes, haproxy, keepalived
from kubemarine.core import flow, summary
from kubemarine.core.cluster import KubernetesCluster
from kubemarine.core.group import NodeGroup
from kubemarine.procedures import install, add_node


def get_active_nodes(node_type: str, cluster: KubernetesCluster) -> NodeGroup:
    all_nodes = cluster.get_nodes_for_removal().having_roles([node_type])
    if all_nodes.is_empty():
        cluster.log.debug("Skipped - no %s to remove" % node_type)
        return all_nodes
    active_nodes = all_nodes.get_online_nodes(True)
    disabled_nodes = all_nodes.exclude_group(active_nodes)
    if active_nodes.is_empty():
        cluster.log.debug("Skipped - %s nodes are inactive: %s" % (node_type, ", ".join(disabled_nodes.nodes)))
        return active_nodes
    if not disabled_nodes.is_empty():
        cluster.log.debug("Partly Skipped - several %s nodes are inactive: %s"
                          % (node_type, ", ".join(disabled_nodes.nodes)))
    return active_nodes


def loadbalancer_remove_haproxy(cluster: KubernetesCluster) -> None:
    nodes = get_active_nodes("balancer", cluster)
    if nodes.is_empty():
        return
    nodes.call(haproxy.disable)


def loadbalancer_remove_keepalived(cluster: KubernetesCluster) -> None:
    nodes = get_active_nodes("keepalived", cluster)
    if nodes.is_empty():
        return
    nodes.call(keepalived.disable)


def remove_kubernetes_nodes(cluster: KubernetesCluster) -> None:
    group = cluster.get_nodes_for_removal().having_roles(['control-plane', 'worker'])

    if group.is_empty():
        cluster.log.debug("No kubernetes nodes to perform")
        return

    group.call(kubernetes.reset_installation_env)
    kubernetes.schedule_running_nodes_report(cluster)


tasks = OrderedDict({
    "loadbalancer": {
        "remove": {
            "haproxy": loadbalancer_remove_haproxy,
            "keepalived": loadbalancer_remove_keepalived
        },
        "configure": {
            "haproxy": install.deploy_loadbalancer_haproxy_configure,
            "keepalived": install.deploy_loadbalancer_keepalived_configure
        },
    },
    "update": {
        "etc_hosts": install.system_prepare_dns_etc_hosts,
        "coredns": install.deploy_coredns,
        "plugins": add_node.redeploy_plugins_if_needed,
    },
    "remove_kubernetes_nodes": remove_kubernetes_nodes,
    "overview": install.overview,
})


cumulative_points = {
    summary.exec_delayed: [
        flow.END_OF_TASKS
    ]
}


class RemoveNodeAction(flow.TasksAction):
    def __init__(self) -> None:
        super().__init__('remove node', tasks,
                         cumulative_points=cumulative_points, recreate_inventory=True)


def create_context(cli_arguments: List[str] = None) -> dict:

    cli_help = '''
    Script for removing node from Kubernetes cluster.

    How to use:

    '''

    parser = flow.new_procedure_parser(cli_help, tasks=tasks)
    context = flow.create_context(parser, cli_arguments, procedure='remove_node')
    return context


def main(cli_arguments: List[str] = None) -> None:
    context = create_context(cli_arguments)
    flow.ActionsFlow([RemoveNodeAction()]).run_flow(context)


if __name__ == '__main__':
    main()



# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import base64
import json
from typing import Callable, Dict, Any, List, Union, Set, Type
from urllib.parse import quote_plus

import yaml
import jinja2

from kubemarine.core import log, utils, errors
from kubemarine.core.proxytypes import (
    Index, Primitive, NodeMapping, ProxyDumper, ProxyJSONEncoder, MutableNode, Node
)

Path = tuple
FILTER = Callable[[str], Any]


class JinjaNode(MutableNode):
    """
    A Node that compiles template strings in the underlying `dict` or `list` on-the-fly during access to its items.
    """

    def __init__(self, delegate: Union[dict, list],
                 *, path: Path, env: 'Environment'):
        super().__init__(delegate)
        self.path = path
        self.env = env

    def _child(self, index: Index, val: Union[list, dict]) -> Node:
        return self._child_type(index)(val, path=self.path + (index,), env=self.env)

    def _child_type(self, _: Index) -> Type['JinjaNode']:
        return JinjaNode

    def _convert(self, index: Index, val: Primitive) -> Primitive:
        if isinstance(val, str) and is_template(val):
            val = self.env.compile_string(val, self.path + (index,))

        return val


class Context(jinja2.runtime.Context):
    """
    An entry point from the jinja templates to the recursively compiled sections of inventory.
    """

    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self.environment: Environment = self.environment

    def resolve_or_missing(self, key: str) -> Any:
        # pylint: disable=protected-access

        v = super().resolve_or_missing(key)

        if v is jinja2.runtime.missing and key in self.environment._recursive_values:
            return self.environment._proxy_values[key]

        return v


class Environment(jinja2.Environment):
    """
    Jinja environment that supports recursive compilation.
    """

    context_class = Context

    def __init__(self, logger: log.EnhancedLogger, recursive_values: dict, *, recursive_extra: Dict[str, Any] = None):
        """
        Instantiate new environment and set default filters.

        :param logger: EnhancedLogger
        :param recursive_values: If templates access to these values, they are automatically compiled if necessary.
        :param recursive_extra: If recursive compilation occurs, these render values are supplied to the template.
        """
        super().__init__()
        self.logger = logger

        self._recursive_values = recursive_values
        self._proxy_values = NodeMapping(self.create_root(self._recursive_values))

        self._compiled: Set[Path] = set()
        self._compiling: List[Path] = []

        self._recursive_extra = {}
        if recursive_extra is not None:
            self._recursive_extra = recursive_extra

        self.policies['json.dumps_function'] = jinja_tojson
        self.filters['toyaml'] = jinja_toyaml

        simple_string_filters: Dict[str, FILTER] = {
            'isipv4': lambda ip: utils.isipv(ip, [4]),
            'isipv6': lambda ip: utils.isipv(ip, [6]),
            'minorversion': utils.minor_version,
            'majorversion': utils.major_version,
            'versionkey': utils.version_key,
            'b64encode': lambda s: base64.b64encode(s.encode()).decode(),
            'b64decode': lambda s: base64.b64decode(s.encode()).decode(),
            'url_quote': quote_plus
        }

        for name, filter_ in simple_string_filters.items():
            def make_filter(n: str, f: FILTER) -> FILTER:
                return lambda s, *args, **kwargs: f(self._check_filter(n, s, *args, *kwargs))

            self.filters[name] = make_filter(name, filter_)

        self.tests['has_role'] = lambda node, role: role in node['roles']
        self.tests['has_roles'] = lambda node, roles: bool(set(node['roles']) & set(roles))

        # we need these filters because rendered cluster.yaml can contain variables like
        # enable: 'true'
        self.filters['is_true'] = utils.strtobool
        self.filters['is_false'] = lambda v: not utils.strtobool(v)

    def create_root(self, delegate: dict) -> Node:
        """
        Create the root wrapper over the recursively compiled container (inventory).

        :param delegate: the root container
        :return: the root wrapper
        """
        return JinjaNode(delegate, path=Path(), env=self)

    def compile_string(self, struct: str, path: Path) -> str:
        """
        Compiles template string at the specified inventory `path`.
        It is called both while going over the inventory,
        and recursively if variables are accessed from templates.

        :param struct: template string
        :param path: path of sections in the inventory
        :return: compiled string
        """
        if path in self._compiled:
            return struct

        if path in self._compiling:
            idx = self._compiling.index(path)
            raise Exception(
                f"Cyclic dynamic variables in inventory{' -> '.join(map(utils.pretty_path, self._compiling[idx:]))}")

        self.logger.verbose("Rendering \"%s\"" % struct)

        self._compiling.append(path)

        try:
            struct = self.from_string(struct).render(self._recursive_extra)
        except errors.BaseKME:
            raise
        except Exception as e:
            raise ValueError(f"Failed to render {struct!r}\nin section {utils.pretty_path(path)}: {e}") from None

        self._compiling.pop()

        self._compiled.add(path)

        self.logger.verbose("\tRendered as \"%s\"" % struct)
        return struct

    def _check_filter(self, filter_: str, struct: str, *args: Any, **kwargs: Any) -> str:
        if args or kwargs:
            raise ValueError(f"Filter {filter_!r} does not support extra arguments")

        if not isinstance(struct, str):
            raise ValueError(f"Filter {filter_!r} can be applied only on string")

        return struct


def jinja_tojson(obj: Any, **kwargs: Any) -> str:
    return json.dumps(obj, cls=ProxyJSONEncoder, **kwargs)


def jinja_toyaml(data: Any) -> str:
    return yaml.dump(data, Dumper=ProxyDumper)


def is_template(struct: str) -> bool:
    return '{{' in struct or '{%' in struct


def compile_node(struct: Union[list, dict], path: List[Union[str, int]], env: Environment) -> Union[list, dict]:
    if isinstance(struct, list):
        for i, v in enumerate(struct):
            struct[i] = compile_object(v, path, i, env)
    else:
        for k, v in struct.items():
            struct[k] = compile_object(v, path, k, env)

    return struct


def compile_object(struct: Union[Primitive, list, dict],
                   path: List[Index], index: Index, env: Environment) -> Union[Primitive, list, dict]:
    path.append(index)
    if isinstance(struct, (list, dict)):
        struct = compile_node(struct, path, env)
    elif isinstance(struct, str) and is_template(struct):
        struct = env.compile_string(struct, Path(path))

    path.pop()
    return struct
